from __future__ import annotations


"""
This module implements lazy evaluation based on weak reference. Meanwhile,
this also introduces functional programming into the project paradigm.

Every child class of `LazyObject` shall define `__slots__`, and all methods
shall be basically sorted in the following way:
- magic methods;
- lazy variables;
- lazy properties;
- private class methods;
- private methods;
- public methods.

All methods decorated by any decorator provided by `Lazy` should be static
methods and be named with underscores appeared on both sides, i.e. `_data_`.
Successive underscores shall not occur, due to the name convension handled by
lazy properties.

The constructor of `LazyDescriptor[T, DataT]` has the following notable
parameters:

- `method: Callable[..., DataT]`
  Defines how the data is calculated through parameters. The returned data
  should be a single element, or a tuple of elements, determined by
  the `plural` flag.

  When `is_property` is false, the method should not take any parameter, and
  returns the initial value for the variable slot.

  The name of each parameter should be concatenated by names of a descriptor
  chain started from the local class, with underscores stripped. For example,
  the name `a__b__c` under class `A` fetches data through the path
  `A._a_._b_._c_`. The fetched data will be an `n`-layer tuple tree, where `n`
  is the number of descriptors with their `plural` flags set to be true.
  Finally, the last descriptor should have `freeze` set true.

- `is_property: bool`
  Lazy.variable: false
  Lazy.volatile: false
  Lazy.property: true

  Determines whether the descriptor behaves as a variable or a property.

  One can call `__set__` of the descriptor on some instance only when:
  - `is_property` is false;
  - the instance is not frozen.

- `plural: bool`
  Lazy.variable: =false
  Lazy.volatile: =false
  Lazy.property: =false

  Determines whether data contains exactly one or arbitrarily many elements.
  When true, `DataT` is specialized as `tuple[T]`; when false, specialized
  as `T`.

- `freeze: bool`
  Lazy.variable: true
  Lazy.volatile: false
  Lazy.property: true

  Determines whether data can be internally modified.

  One shall note that actually no freezing procedure is taken in runtime for
  better performance. Internally modifying data bound to a descriptor with
  `freeze` set true may result in unexpected results.

- `deepcopy: bool`
  Lazy.variable: false
  Lazy.volatile: =true
  Lazy.property: false

  Determines how data in the descriptor is copied when calling
  `LazyObject.copy` from the parent object. Does not take effect when
  `is_property` is true, since only data in variable descriptors are copied.

- `cache_capacity: int`
  Lazy.variable: 1
  Lazy.volatile: 0
  Lazy.property: =128

  Determines the capacity of the lru cache of parameters-data pairs generated
  from `method`.

Descriptor overriding is allowed. The overriding descriptor should match the
overridden one in `plural`, `freeze` flags (a change in `freeze` from false to
true is allowed). Furthermore, the element type (the specialization of type
variable `T`) should be consistent between descriptors.
"""


from typing import (
    Callable,
    Literal,
    Never,
    Self,
    overload
)

from .lazy_descriptor import LazyDescriptor


class Lazy:
    __slots__ = ()

    def __new__(
        cls: type[Self]
    ) -> Never:
        raise TypeError

    @overload
    @classmethod
    def variable[T](
        cls: type[Self],
        *,
        plural: Literal[False] = False
    ) -> Callable[[Callable[[], T]], LazyDescriptor[T, T]]: ...

    @overload
    @classmethod
    def variable[T](
        cls: type[Self],
        *,
        plural: Literal[True]
    ) -> Callable[[Callable[[], tuple[T, ...]]], LazyDescriptor[T, tuple[T, ...]]]: ...

    @classmethod
    def variable(
        cls: type[Self],
        *,
        plural: bool = False
    ) -> Callable[[Callable], LazyDescriptor]:

        def result(
            method: Callable
        ) -> LazyDescriptor:
            assert isinstance(method, staticmethod)
            return LazyDescriptor(
                method=method.__func__,
                is_property=False,
                plural=plural,
                freeze=True,
                deepcopy=False,
                cache_capacity=1
            )

        return result

    @overload
    @classmethod
    def volatile[T](
        cls: type[Self],
        *,
        plural: Literal[False] = False,
        deepcopy: bool = True
    ) -> Callable[[Callable[[], T]], LazyDescriptor[T, T]]: ...

    @overload
    @classmethod
    def volatile[T](
        cls: type[Self],
        *,
        plural: Literal[True],
        deepcopy: bool = True
    ) -> Callable[[Callable[[], tuple[T, ...]]], LazyDescriptor[T, tuple[T, ...]]]: ...

    @classmethod
    def volatile(
        cls: type[Self],
        *,
        plural: bool = False,
        deepcopy: bool = True
    ) -> Callable[[Callable], LazyDescriptor]:

        def result(
            method: Callable
        ) -> LazyDescriptor:
            assert isinstance(method, staticmethod)
            return LazyDescriptor(
                method=method.__func__,
                is_property=False,
                plural=plural,
                freeze=False,
                deepcopy=deepcopy,
                cache_capacity=0
            )

        return result

    @overload
    @classmethod
    def property[T](
        cls: type[Self],
        *,
        plural: Literal[False] = False,
        cache_capacity: int = 128
    ) -> Callable[[Callable[..., T]], LazyDescriptor[T, T]]: ...

    @overload
    @classmethod
    def property[T](
        cls: type[Self],
        *,
        plural: Literal[True],
        cache_capacity: int = 128
    ) -> Callable[[Callable[..., tuple[T, ...]]], LazyDescriptor[T, tuple[T, ...]]]: ...

    @classmethod
    def property(
        cls: type[Self],
        *,
        plural: bool = False,
        cache_capacity: int = 128
    ) -> Callable[[Callable], LazyDescriptor]:

        def result(
            method: Callable
        ) -> LazyDescriptor:
            assert isinstance(method, staticmethod)
            return LazyDescriptor(
                method=method.__func__,
                is_property=True,
                plural=plural,
                freeze=True,
                deepcopy=False,
                cache_capacity=cache_capacity
            )

        return result



from __future__ import annotations


__all__ = (
    "ORIGIN",
    "RIGHT",
    "LEFT",
    "UP",
    "DOWN",
    "OUT",
    "IN",
    "X_AXIS",
    "Y_AXIS",
    "Z_AXIS",
    "UR",
    "UL",
    "DL",
    "DR",
    "PI",
    "TAU",
    "DEGREES"
)


import numpy as np

from .custom_typing import NP_3f8


ORIGIN: NP_3f8 = np.array((0.0, 0.0, 0.0))
RIGHT: NP_3f8 = np.array((1.0, 0.0, 0.0))
LEFT: NP_3f8 = np.array((-1.0, 0.0, 0.0))
UP: NP_3f8 = np.array((0.0, 1.0, 0.0))
DOWN: NP_3f8 = np.array((0.0, -1.0, 0.0))
OUT: NP_3f8 = np.array((0.0, 0.0, 1.0))
IN: NP_3f8 = np.array((0.0, 0.0, -1.0))
X_AXIS: NP_3f8 = np.array((1.0, 0.0, 0.0))
Y_AXIS: NP_3f8 = np.array((0.0, 1.0, 0.0))
Z_AXIS: NP_3f8 = np.array((0.0, 0.0, 1.0))

UR: NP_3f8 = UP + RIGHT
UL: NP_3f8 = UP + LEFT
DL: NP_3f8 = DOWN + LEFT
DR: NP_3f8 = DOWN + RIGHT

PI: float = np.pi
TAU: float = PI * 2.0
DEGREES: float = PI / 180.0



from __future__ import annotations


from contextlib import contextmanager
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Iterator,
    Self
)

if TYPE_CHECKING:
    from .config import Config
    from .context import Context
    from .logger import Logger
    from .renderer import Renderer
    from .scene import Scene
    from .timer import Timer
    from .window import Window


class Toplevel:
    __slots__ = ()

    _config: ClassVar[Config | None] = None
    _timer: ClassVar[Timer | None] = None
    _logger: ClassVar[Logger | None] = None
    _window: ClassVar[Window | None] = None
    _context: ClassVar[Context | None] = None
    _renderer: ClassVar[Renderer | None] = None
    _scene: ClassVar[Scene | None] = None

    @classmethod
    def _get_config(
        cls: type[Self]
    ) -> Config:
        assert (config := cls._config) is not None
        return config

    @classmethod
    def _get_timer(
        cls: type[Self]
    ) -> Timer:
        assert (timer := cls._timer) is not None
        return timer

    @classmethod
    def _get_logger(
        cls: type[Self]
    ) -> Logger:
        assert (logger := cls._logger) is not None
        return logger

    @classmethod
    def _get_window(
        cls: type[Self]
    ) -> Window:
        assert (window := cls._window) is not None
        return window

    @classmethod
    def _get_context(
        cls: type[Self]
    ) -> Context:
        assert (context := cls._context) is not None
        return context

    @classmethod
    def _get_renderer(
        cls: type[Self]
    ) -> Renderer:
        assert (renderer := cls._renderer) is not None
        return renderer

    @classmethod
    def _get_scene(
        cls: type[Self]
    ) -> Scene:
        assert (scene := cls._scene) is not None
        return scene

    @classmethod
    def start_livestream(
        cls: type[Self]
    ) -> None:
        cls._get_renderer().start_livestream()

    @classmethod
    def stop_livestream(
        cls: type[Self]
    ) -> None:
        cls._get_renderer().stop_livestream()

    @classmethod
    def start_recording(
        cls: type[Self],
        filename: str | None = None
    ) -> None:
        cls._get_renderer().start_recording(filename)

    @classmethod
    def stop_recording(
        cls: type[Self],
        filename: str | None = None
    ) -> None:
        cls._get_renderer().stop_recording(filename)

    @classmethod
    def snapshot(
        cls: type[Self],
        filename: str | None = None
    ) -> None:
        cls._get_renderer().snapshot(filename)

    @classmethod
    @contextmanager
    def livestream(
        cls: type[Self]
    ) -> Iterator[None]:
        cls.start_livestream()
        yield
        cls.stop_livestream()

    @classmethod
    @contextmanager
    def recording(
        cls: type[Self],
        filename: str | None = None
    ) -> Iterator[None]:
        cls.start_recording(filename)
        yield
        cls.stop_recording(filename)



from __future__ import annotations


from typing import Self

import numpy as np

from ..constants.custom_typing import NP_3f8
from ..lazy.lazy import Lazy
from ..rendering.buffers.uniform_block_buffer import UniformBlockBuffer
from .animatable.animatable import (
    Animatable,
    AnimatableActions
)
from .lights.ambient_light import AmbientLight
from .lights.point_light import PointLight


class Lighting(Animatable):
    __slots__ = ()

    def __init__(
        self: Self,
        *lights: AmbientLight | PointLight
    ) -> None:
        super().__init__()
        ambient_lights: list[AmbientLight] = []
        point_lights: list[PointLight] = []
        for light in lights:
            if isinstance(light, AmbientLight):
                ambient_lights.append(light)
            else:
                point_lights.append(light)
        self._ambient_lights_ = tuple(ambient_lights)
        self._point_lights_ = tuple(point_lights)

    @AnimatableActions.interpolate.register_descriptor()
    @Lazy.volatile(plural=True)
    @staticmethod
    def _ambient_lights_() -> tuple[AmbientLight, ...]:
        return ()

    @AnimatableActions.interpolate.register_descriptor()
    @Lazy.volatile(plural=True)
    @staticmethod
    def _point_lights_() -> tuple[PointLight, ...]:
        return ()

    @Lazy.property()
    @staticmethod
    def _lighting_uniform_block_buffer_(
        ambient_lights__color__array: tuple[NP_3f8, ...],
        point_lights__color__array: tuple[NP_3f8, ...],
        point_lights__position: tuple[NP_3f8, ...]
    ) -> UniformBlockBuffer:
        return UniformBlockBuffer(
            name="ub_lighting",
            field_declarations=(
                "AmbientLight u_ambient_lights[NUM_U_AMBIENT_LIGHTS]",
                "PointLight u_point_lights[NUM_U_POINT_LIGHTS]"
            ),
            structs={
                "AmbientLight": (
                    "vec3 color",
                ),
                "PointLight": (
                    "vec3 position",
                    "vec3 color"
                )
            },
            data_dict={
                "u_ambient_lights.color": np.fromiter(
                    ambient_lights__color__array,
                    dtype=np.dtype((np.float64, (3,)))
                ),
                "u_point_lights.position": np.fromiter(
                    point_lights__position,
                    dtype=np.dtype((np.float64, (3,)))
                ),
                "u_point_lights.color": np.fromiter(
                    point_lights__color__array,
                    dtype=np.dtype((np.float64, (3,)))
                )
            },
            array_lens={
                "NUM_U_AMBIENT_LIGHTS": len(ambient_lights__color__array),
                "NUM_U_POINT_LIGHTS": len(point_lights__color__array)
            }
        )



from __future__ import annotations


import itertools
import weakref
from typing import (
    Iterator,
    Self
)

from ..animatables.camera import Camera
from ..animatables.lighting import Lighting
from ..animatables.model import Model
from ..lazy.lazy import Lazy
from ..rendering.vertex_array import VertexArray
from ..toplevel.toplevel import Toplevel


class Mobject(Model):
    __slots__ = (
        "__weakref__",
        "_children",
        "_proper_descendants",
        "_parents",
        "_proper_ancestors"
    )

    def __init__(
        self: Self
    ) -> None:
        super().__init__()
        self._children: list[Mobject] = []
        self._proper_descendants: list[Mobject] = []
        self._parents: weakref.WeakSet[Mobject] = weakref.WeakSet()
        self._proper_ancestors: weakref.WeakSet[Mobject] = weakref.WeakSet()

    def __iter__(
        self: Self
    ) -> Iterator[Mobject]:
        yield from self._children

    # family matters
    # These methods implement a DAG (directed acyclic graph).

    @classmethod
    def _refresh_families(
        cls: type[Self],
        *mobjects: Mobject
    ) -> None:

        def iter_descendants_by_children(
            mobject: Mobject
        ) -> Iterator[Mobject]:
            yield mobject
            for child in mobject._children:
                yield from iter_descendants_by_children(child)

        def iter_ancestors_by_parents(
            mobject: Mobject
        ) -> Iterator[Mobject]:
            yield mobject
            for parent in mobject._parents:
                yield from iter_ancestors_by_parents(parent)

        for proper_ancestor in dict.fromkeys(itertools.chain.from_iterable(
            iter_ancestors_by_parents(mobject)
            for mobject in mobjects
        )):
            proper_descendants = dict.fromkeys(itertools.chain.from_iterable(
                iter_descendants_by_children(child)
                for child in proper_ancestor._children
            ))
            proper_ancestor._proper_descendants.clear()
            proper_ancestor._proper_descendants.extend(proper_descendants)
            proper_ancestor._proper_siblings_ = tuple(proper_descendants)

        for proper_descendant in dict.fromkeys(itertools.chain.from_iterable(
            iter_descendants_by_children(mobject)
            for mobject in mobjects
        )):
            proper_ancestors = dict.fromkeys(itertools.chain.from_iterable(
                iter_descendants_by_children(parent)
                for parent in proper_descendant._parents
            ))
            proper_descendant._proper_ancestors.clear()
            proper_descendant._proper_ancestors.update(proper_ancestors)

    def iter_children(
        self: Self
    ) -> Iterator[Mobject]:
        yield from self._children

    def iter_parents(
        self: Self
    ) -> Iterator[Mobject]:
        yield from self._parents

    def iter_descendants(
        self: Self,
        *,
        broadcast: bool = True
    ) -> Iterator[Mobject]:
        yield self
        if broadcast:
            yield from self._proper_descendants

    def iter_ancestors(
        self: Self,
        *,
        broadcast: bool = True
    ) -> Iterator[Mobject]:
        yield self
        if broadcast:
            yield from self._proper_ancestors

    def add(
        self: Self,
        *mobjects: Mobject
    ) -> Self:
        if (invalid_mobjects := tuple(
            mobject for mobject in mobjects
            if mobject in self.iter_ancestors()
        )):
            raise ValueError(f"Circular relationship occurred when adding {invalid_mobjects} to {self}")
        for mobject in dict.fromkeys(mobjects):
            if mobject in self._children:
                continue
            self._children.append(mobject)
            mobject._parents.add(self)
        type(self)._refresh_families(self, *mobjects)
        return self

    def discard(
        self: Self,
        *mobjects: Mobject
    ) -> Self:
        for mobject in dict.fromkeys(mobjects):
            if mobject not in self._children:
                continue
            self._children.remove(mobject)
            mobject._parents.remove(self)
        type(self)._refresh_families(self, *mobjects)
        return self

    def clear(
        self: Self
    ) -> Self:
        self.discard(*self.iter_children())
        return self

    def copy(
        self: Self
    ) -> Self:
        # Copy all descendants. The result is not bound to any mobject.
        result = super().copy()
        descendants: list[Mobject] = [self, *(
            descendant for descendant in self._proper_siblings_
            if isinstance(descendant, Mobject)
        )]
        descendants_copy: list[Mobject] = [result, *(
            descendant_copy for descendant_copy in result._proper_siblings_
            if isinstance(descendant_copy, Mobject)
        )]

        for descendant, descendant_copy in zip(descendants, descendants_copy, strict=True):
            descendant_copy._children = [
                descendants_copy[descendants.index(child)]
                for child in descendant._children
            ]
            descendant_copy._proper_descendants = []
            descendant_copy._parents = weakref.WeakSet(
                descendants_copy[descendants.index(parent)]
                for parent in descendant._parents
                if parent in descendants
            )
            descendant_copy._proper_ancestors = weakref.WeakSet()

        type(self)._refresh_families(*descendants_copy)
        return result

    # render

    @Lazy.volatile(deepcopy=False)
    @staticmethod
    def _camera_() -> Camera:
        return Toplevel._get_scene()._camera

    @Lazy.volatile(deepcopy=False)
    @staticmethod
    def _lighting_() -> Lighting:
        return Toplevel._get_scene()._lighting

    def _iter_vertex_arrays(
        self: Self
    ) -> Iterator[VertexArray]:
        yield from ()

    def bind_camera(
        self: Self,
        camera: Camera,
        *,
        broadcast: bool = True,
    ) -> Self:
        for mobject in self.iter_descendants(broadcast=broadcast):
            mobject._camera_ = camera
        return self

    def bind_lighting(
        self: Self,
        lighting: Lighting,
        *,
        broadcast: bool = True,
    ) -> Self:
        for mobject in self.iter_descendants(broadcast=broadcast):
            mobject._lighting_ = lighting
        return self



from __future__ import annotations


from typing import (
    Callable,
    Self
)

import numpy as np

from ...animatables.mesh import Mesh
from ...constants.custom_typing import (
    NP_x2f8,
    NP_x3f8
)
from .mesh_mobject import MeshMobject


class ParametricSurface(MeshMobject):
    __slots__ = ()

    def __init__(
        self: Self,
        *,
        func: Callable[[NP_x2f8], NP_x3f8],
        normal_func: Callable[[NP_x2f8], NP_x3f8],
        u_range: tuple[float, float],
        v_range: tuple[float, float],
        resolution: tuple[int, int] = (128, 128)
    ) -> None:
        u_start, u_stop = u_range
        v_start, v_stop = v_range
        u_len = resolution[0] + 1
        v_len = resolution[1] + 1
        index_grid = np.mgrid[0:u_len, 0:v_len]
        ne = index_grid[:, +1:, +1:]
        nw = index_grid[:, :-1, +1:]
        sw = index_grid[:, :-1, :-1]
        se = index_grid[:, +1:, :-1]
        faces = np.ravel_multi_index(
            tuple(np.stack((se, sw, ne, sw, nw, ne), axis=3)),
            (u_len, v_len)
        ).reshape((-1, 3))

        uvs = np.stack(np.meshgrid(
            np.linspace(0.0, 1.0, u_len),
            np.linspace(0.0, 1.0, v_len),
            indexing="ij"
        ), 2).reshape((-1, 2))
        samples: NP_x2f8 = np.stack(np.meshgrid(
            np.linspace(u_start, u_stop, u_len),
            np.linspace(v_start, v_stop, v_len),
            indexing="ij"
        ), 2).reshape((-1, 2))
        positions = func(samples)
        normals = normal_func(samples)

        super().__init__(Mesh(
            positions=positions,
            normals=normals,
            uvs=uvs,
            faces=faces
        ))



from __future__ import annotations


import pathlib
import re
from typing import (
    Iterator,
    Self,
    TypedDict
)

import attrs

from ...animatables.arrays.animatable_color import AnimatableColor
from ...constants.custom_typing import ColorType
from ...toplevel.toplevel import Toplevel
from .string_mobject import (
    BalancedCommandInfo,
    CommandInfo,
    StandaloneCommandInfo,
    StringMobjectIO,
    StringMobjectInput,
    StringMobjectKwargs
)


class LatexAttributes(TypedDict, total=False):
    color: str


@attrs.frozen(kw_only=True)
class LatexStringMobjectInput(StringMobjectInput[LatexAttributes]):
    color: ColorType = attrs.field(factory=lambda: Toplevel._get_config().default_color)


class LatexStringMobjectKwargs(StringMobjectKwargs[LatexAttributes], total=False):
    color: ColorType


class LatexStringMobjectIO[LatexStringMobjectInputT: LatexStringMobjectInput](StringMobjectIO[LatexAttributes, LatexStringMobjectInputT]):
    __slots__ = ()

    @classmethod
    def _iter_global_span_attributes(
        cls: type[Self],
        input_data: LatexStringMobjectInputT,
        temp_path: pathlib.Path
    ) -> Iterator[LatexAttributes]:
        yield LatexAttributes(
            color=AnimatableColor._color_to_hex(input_data.color)
        )
        yield from super()._iter_global_span_attributes(input_data, temp_path)

    @classmethod
    def _get_empty_attributes(
        cls: type[Self]
    ) -> LatexAttributes:
        return LatexAttributes()

    @classmethod
    def _get_command_pair(
        cls: type[Self],
        attributes: LatexAttributes
    ) -> tuple[str, str]:
        if (color_hex := attributes.get("color")) is None:
            return "", ""
        match = re.fullmatch(r"#([0-9A-F]{2})([0-9A-F]{2})([0-9A-F]{2})", color_hex, flags=re.IGNORECASE)
        assert match is not None
        return f"{{\\color[RGB]{{{", ".join(
            f"{int(match.group(index), 16)}"
            for index in range(1, 4)
        )}}}{{", f"}}}}"

    @classmethod
    def _convert_attributes_for_labelling(
        cls: type[Self],
        attributes: LatexAttributes,
        label: int | None
    ) -> LatexAttributes:
        result = LatexAttributes(attributes)
        for key in (
            "color",
        ):
            if key in result:
                result.pop(key)

        if label is not None:
            result["color"] = f"#{label:06x}"
        return result

    @classmethod
    def _iter_command_infos(
        cls: type[Self],
        string: str
    ) -> Iterator[CommandInfo[LatexAttributes]]:
        pattern = re.compile(r"""
            (?P<command>\\(?:[a-zA-Z]+|.))
            |(?P<open>{+)
            |(?P<close>}+)
        """, flags=re.VERBOSE | re.DOTALL)
        open_stack: list[tuple[int, int]] = []
        for match in pattern.finditer(string):
            if not match.group("close"):
                if not match.group("open"):
                    yield StandaloneCommandInfo(match)
                    continue
                open_stack.append(match.span())
                continue
            close_start, close_stop = match.span()
            while True:
                if not open_stack:
                    raise ValueError("Missing '{' inserted")
                open_start, open_stop = open_stack.pop()
                width = min(open_stop - open_start, close_stop - close_start)
                assert (open_match_obj := pattern.fullmatch(string, pos=open_stop - width, endpos=open_stop)) is not None
                assert (close_match_obj := pattern.fullmatch(string, pos=close_start, endpos=close_start + width)) is not None
                yield BalancedCommandInfo(
                    attributes=LatexAttributes(),
                    isolated=width >= 2,
                    open_match_obj=open_match_obj,
                    close_match_obj=close_match_obj
                )
                open_stop -= width
                close_start += width
                if close_start < close_stop:
                    continue
                if open_start < open_stop:
                    open_stack.append((open_start, open_stop))
                break
        if open_stack:
            raise ValueError("Missing '}' inserted")



from __future__ import annotations


import pathlib
import re
import subprocess
from typing import (
    Self,
    Unpack
)

import attrs

from ...constants.custom_typing import AlignmentType
from ...toplevel.toplevel import Toplevel
from .latex_string_mobject import (
    LatexStringMobjectIO,
    LatexStringMobjectInput,
    LatexStringMobjectKwargs
)
from .string_mobject import StringMobject


@attrs.frozen(kw_only=True)
class TexInput(LatexStringMobjectInput):
    alignment: AlignmentType = attrs.field(factory=lambda: Toplevel._get_config().tex_alignment)
    compiler: str = attrs.field(factory=lambda: Toplevel._get_config().tex_compiler)
    preambles: tuple[str, ...] = attrs.field(factory=lambda: Toplevel._get_config().tex_preambles)


class TexKwargs(LatexStringMobjectKwargs, total=False):
    alignment: AlignmentType
    compiler: str
    preambles: tuple[str, ...]


class TexIO[TexInputT: TexInput](LatexStringMobjectIO[TexInputT]):
    __slots__ = ()

    @classmethod
    def _get_subdir_name(
        cls: type[Self]
    ) -> str:
        return "tex"

    @classmethod
    def _create_svg(
        cls: type[Self],
        content: str,
        input_data: TexInputT,
        svg_path: pathlib.Path
    ) -> None:
        match input_data.compiler:
            case "latex":
                compiler_args = ("latex",)
                dvi_suffix = ".dvi"
            case "xelatex":
                compiler_args = ("xelatex", "-no-pdf")
                dvi_suffix = ".xdv"
            case _:
                raise ValueError(f"Compiler '{input_data.compiler}' is not implemented")

        match input_data.alignment:
            case "left":
                alignment_command = "\\flushleft"
            case "center":
                alignment_command = "\\centering"
            case "right":
                alignment_command = "\\flushright"

        full_content = "\n".join((
            "\\documentclass[preview]{standalone}",
            *input_data.preambles,
            "\\begin{document}",
            alignment_command,
            content,
            "\\end{document}"
        )) + "\n"

        tex_path = svg_path.with_suffix(".tex")
        tex_path.write_text(full_content, encoding="utf-8")

        try:
            if subprocess.run((
                *compiler_args,
                tex_path,
                "-interaction", "batchmode",
                "-halt-on-error",
                "-output-directory", svg_path.parent
            ), stdout=subprocess.DEVNULL).returncode or subprocess.run((
                "dvisvgm",
                svg_path.with_suffix(dvi_suffix),
                "-n",
                "-v", "0",
                "-o", svg_path
            ), stdout=subprocess.DEVNULL).returncode:
                error = OSError(f"TexIO: Failed to execute latex command")
                log_text = svg_path.with_suffix(".log").read_text(encoding="utf-8")
                if (error_match_obj := re.search(r"(?<=\n! ).*", log_text)) is not None:
                    error.add_note(f"LaTeX error: {error_match_obj.group()}")
                raise error

        finally:
            for suffix in (".tex", dvi_suffix, ".log", ".aux"):
                svg_path.with_suffix(suffix).unlink(missing_ok=True)

    @classmethod
    def _get_adjustment_scale(
        cls: type[Self]
    ) -> float:
        return 0.04579


class Tex(StringMobject):
    __slots__ = ()

    def __init__(
        self: Self,
        string: str,
        **kwargs: Unpack[TexKwargs]
    ) -> None:
        super().__init__(TexIO.get(TexInput(string=string, **kwargs)))



from __future__ import annotations


from typing import Self

from ..animatables.arrays.animatable_color import AnimatableColor
from ..animatables.lights.ambient_light import AmbientLight
from ..animatables.camera import Camera
from ..animatables.lighting import Lighting
from ..constants.custom_typing import ColorType
from ..mobjects.mobject import Mobject
from ..timelines.timeline import Timeline
from .toplevel import Toplevel


class Scene(Timeline):
    __slots__ = (
        "_camera",
        "_lighting",
        "_background_color",
        "_background_opacity",
        "_root_mobject",
        "_scene_time"
    )

    def __init__(
        self: Self
    ) -> None:
        super().__init__()
        self._camera: Camera = Camera()
        self._lighting: Lighting = Lighting(AmbientLight())
        self._background_color: tuple[float, float, float] = (0.0, 0.0, 0.0)
        self._background_opacity: float = 0.0
        self._root_mobject: Mobject = Mobject()
        self._scene_time: float = 0.0
        self.set(
            background_color=Toplevel._get_config().background_color,
            background_opacity=Toplevel._get_config().background_opacity
        )

    def _run(
        self: Self
    ) -> None:
        self.schedule(parent_absolute_rate=lambda: Toplevel._get_scene()._scene_time)
        for scene_time in Toplevel._get_timer().frame_clock():
            self._scene_time = scene_time
            Toplevel._get_window()._pyglet_window.dispatch_events()
            self._progress()
            Toplevel._get_window().clear_event_info_queue()
            Toplevel._get_renderer().process_frame()
            if self.terminated():
                break

    def run(
        self: Self
    ) -> None:
        Toplevel._scene = self
        try:
            self._run()
        except KeyboardInterrupt:
            pass
        Toplevel._scene = None

    # Shortcut access to root mobject.

    def add(
        self: Self,
        *mobjects: Mobject
    ) -> Self:
        self._root_mobject.add(*mobjects)
        return self

    def discard(
        self: Self,
        *mobjects: Mobject
    ) -> Self:
        self._root_mobject.discard(*mobjects)
        return self

    def clear(
        self: Self
    ) -> Self:
        self._root_mobject.clear()
        return self

    @property
    def camera(
        self: Self
    ) -> Camera:
        return self._camera

    @property
    def lighting(
        self: Self
    ) -> Lighting:
        return self._lighting

    def bind_camera(
        self: Self,
        camera: Camera,
        *,
        broadcast: bool = True
    ) -> Self:
        self._camera = camera
        self._root_mobject.bind_camera(camera, broadcast=broadcast)
        return self

    def bind_lighting(
        self: Self,
        lighting: Lighting,
        *,
        broadcast: bool = True
    ) -> Self:
        self._lighting = lighting
        self._root_mobject.bind_lighting(lighting, broadcast=broadcast)
        return self

    def set(
        self: Self,
        camera: Camera | None = None,
        lighting: Lighting | None = None,
        background_color: ColorType | None = None,
        background_opacity: float | None = None,
        *,
        broadcast: bool = True
    ) -> Self:
        if camera is not None:
            self.bind_camera(camera, broadcast=broadcast)
        if lighting is not None:
            self.bind_lighting(lighting, broadcast=broadcast)
        if background_color is not None:
            red, green, blue = AnimatableColor._color_to_array(background_color)
            self._background_color = (float(red), float(green), float(blue))
        if background_opacity is not None:
            self._background_opacity = background_opacity
        return self



from __future__ import annotations


import copy
import functools
import inspect
from abc import ABC
from enum import Enum
from typing import (
    Any,
    Callable,
    ClassVar,
    Hashable,
    Never,
    Self,
    TypeAliasType,
    TypeVar
)

import attrs
import numpy as np

from .lazy_descriptor import LazyDescriptor
from .lazy_slot import LazySlot


@attrs.frozen(kw_only=True)
class AnnotationRecord:
    annotation: Any
    annotation_cls: type
    type_params_cls: type

    def specialize(
        self: Self,
        annotation_record: AnnotationRecord
    ) -> AnnotationRecord:
        assert self.type_params_cls is annotation_record.annotation_cls
        type_params = self.type_params_cls.__type_params__
        specializations = getattr(annotation_record.annotation, "__args__", ())
        assert len(type_params) == len(specializations)
        annotation = self.annotation
        if annotation in type_params:
            new_annotation = specializations[type_params.index(annotation)]
        elif (parameters := getattr(annotation, "__parameters__", ())):
            new_annotation = annotation[*(
                specializations[type_params.index(type_param)]
                for type_param in parameters
            )]
        else:
            new_annotation = annotation
        return AnnotationRecord(
            annotation=new_annotation,
            annotation_cls=self.annotation_cls,
            type_params_cls=annotation_record.type_params_cls
        )

    @classmethod
    def from_override(
        cls: type[Self],
        override_cls: type,
        base_cls: type
    ) -> AnnotationRecord:
        return AnnotationRecord(
            annotation=base_cls if not base_cls.__type_params__ else override_cls.__orig_bases__[0],
            annotation_cls=base_cls,
            type_params_cls=override_cls
        )

    @classmethod
    def from_root(
        cls: type[Self],
        root_cls: type
    ) -> AnnotationRecord:
        return AnnotationRecord(
            annotation=root_cls if not root_cls.__type_params__ else root_cls[*root_cls.__type_params__],
            annotation_cls=root_cls,
            type_params_cls=root_cls
        )

    @classmethod
    def get_cls_from_annotation(
        cls: type[Self],
        annotation: Any
    ) -> type:
        if isinstance(annotation, TypeVar):
            annotation = annotation.__bound__
        elif isinstance(annotation, TypeAliasType):
            annotation = annotation.__value__
        annotation = getattr(annotation, "__origin__", annotation)
        return annotation if isinstance(annotation, type) else object

    @classmethod
    def extract_element_annotation(
        cls: type[Self],
        annotation: Any,
        plural: bool
    ) -> Any:
        if not plural:
            return annotation
        assert annotation.__origin__ is tuple
        element_annotation, ellipsis = annotation.__args__
        assert ellipsis is ...
        return element_annotation


class LazyObject(ABC):
    __slots__ = ("_lazy_slots",)

    _lazy_descriptors: tuple[LazyDescriptor, ...] = ()
    _annotated_lazy_descriptors: dict[str, tuple[AnnotationRecord, LazyDescriptor]] = {}
    _lazy_slots_cls: ClassVar[type] = object
    _slot_names: ClassVar[tuple[str, ...]] = ()

    def __init_subclass__(
        cls: type[Self]
    ) -> None:
        super().__init_subclass__()

        annotated_lazy_descriptors = {
            name: (annotation_record.specialize(AnnotationRecord.from_override(cls, base)), descriptor)
            for base in reversed(cls.__bases__)
            if issubclass(base, LazyObject)
            for name, (annotation_record, descriptor) in base._annotated_lazy_descriptors.items()
        }
        cls._annotated_lazy_descriptors = annotated_lazy_descriptors

        root_annotation_record = AnnotationRecord.from_root(cls)
        for name, descriptor in cls.__dict__.items():
            if not isinstance(descriptor, LazyDescriptor):
                continue

            signature = inspect.signature(
                descriptor._method,
                locals={
                    cls.__name__: cls,
                    **{
                        str(type_param): type_param
                        for type_param in cls.__type_params__
                    }
                },
                eval_str=True
            )
            element_annotation = AnnotationRecord.extract_element_annotation(
                annotation=signature.return_annotation,
                plural=descriptor._plural
            )
            element_annotation_cls = AnnotationRecord.get_cls_from_annotation(element_annotation)

            assert name.startswith("_") and name.endswith("_") and "__" not in name
            descriptor._name = name
            parameter_name_chains = tuple(
                tuple(f"_{name_body}_" for name_body in parameter_name.split("__"))
                for parameter_name in signature.parameters
            )
            descriptor._parameter_name_chains = parameter_name_chains
            assert descriptor._is_property or not parameter_name_chains

            descriptor._decomposer = Implementations.decomposers.fetch(descriptor._plural)
            descriptor._composer = Implementations.composers.fetch(descriptor._plural)
            descriptor._hasher = (
                Implementations.hashers.fetch(element_annotation_cls if descriptor._freeze else object)
            )

            if (typed_overridden_descriptor := annotated_lazy_descriptors.get(name)) is not None:
                annotation_record, overridden_descriptor = typed_overridden_descriptor
                assert overridden_descriptor._plural is descriptor._plural
                assert not overridden_descriptor._freeze or descriptor._freeze
                assert annotation_record.annotation_cls is element_annotation_cls
                assert annotation_record.annotation == element_annotation

            annotated_lazy_descriptors[name] = (AnnotationRecord(
                annotation=element_annotation,
                annotation_cls=element_annotation_cls,
                type_params_cls=cls
            ), descriptor)

            for parameter_name_chain, parameter in zip(
                parameter_name_chains, signature.parameters.values(), strict=True
            ):
                last_descriptor_freeze: bool = False
                annotation_record = root_annotation_record
                provided_annotation = parameter.annotation
                for name_segment in parameter_name_chain:
                    assert issubclass(element_cls := annotation_record.annotation_cls, LazyObject)
                    parameter_annotation_record, parameter_descriptor = element_cls._annotated_lazy_descriptors[name_segment]
                    last_descriptor_freeze = parameter_descriptor._freeze
                    annotation_record = parameter_annotation_record.specialize(annotation_record)
                    provided_annotation = AnnotationRecord.extract_element_annotation(
                        annotation=provided_annotation,
                        plural=parameter_descriptor._plural
                    )
                assert last_descriptor_freeze
                assert annotation_record.annotation == provided_annotation
                assert parameter.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD
                assert parameter.default is inspect.Parameter.empty

        cls._lazy_descriptors = tuple(descriptor for _, descriptor in annotated_lazy_descriptors.values())
        cls._lazy_slots_cls = attrs.make_class(
            name=f"__{cls.__name__}",
            attrs={
                name: attrs.field(factory=functools.partial(LazySlot, descriptor))
                for name, (_, descriptor) in annotated_lazy_descriptors.items()
            },
            slots=True,
            frozen=True
        )
        cls._slot_names = tuple(
            slot_name
            for base in reversed(cls.__mro__)
            if issubclass(base, LazyObject)
            for slot_name in base.__slots__
            if not slot_name.startswith("__")
            and slot_name != "_lazy_slots"
        )

    def __init__(
        self: Self
    ) -> None:
        super().__init__()
        self._lazy_slots: object = type(self)._lazy_slots_cls()

    def __copy__(
        self: Self
    ) -> Self:
        return self.copy()

    def _get_lazy_slot(
        self: Self,
        name: str
    ) -> LazySlot:
        return self._lazy_slots.__getattribute__(name)

    def copy(
        self: Self
    ) -> Self:
        cls = type(self)
        result = cls.__new__(cls)
        for slot_name in cls._slot_names:
            result.__setattr__(slot_name, copy.copy(self.__getattribute__(slot_name)))
        result._lazy_slots = cls._lazy_slots_cls()
        for descriptor in cls._lazy_descriptors:
            if descriptor._is_property:
                continue
            elements = descriptor.get_elements(self)
            if descriptor._deepcopy:
                elements = tuple(copy.copy(element) for element in elements)
            descriptor.set_elements(result, elements)
        return result


class ImplementationRegistry[K: Hashable, V]:
    __slots__ = (
        "_registry",
        "_match_key"
    )

    def __init__(
        self: Self,
        match_key: Callable[[K, K], bool]
    ) -> None:
        super().__init__()
        self._registry: dict[K, V] = {}
        self._match_key: Callable[[K, K], bool] = match_key

    def register(
        self: Self,
        key: K
    ) -> Callable[[V], V]:

        def result(
            value: V
        ) -> V:
            self._registry[key] = value
            return value

        return result

    def fetch(
        self: Self,
        key: K
    ) -> V:
        for registered_key, value in self._registry.items():
            if self._match_key(key, registered_key):
                return value
        raise NotImplementedError(key)


class Implementations:
    __slots__ = ()

    decomposers: ClassVar[ImplementationRegistry[bool, Callable[[Any], tuple[Any, ...]]]] = ImplementationRegistry(bool.__eq__)
    composers: ClassVar[ImplementationRegistry[bool, Callable[[tuple[Any, ...]], Any]]] = ImplementationRegistry(bool.__eq__)
    hashers: ClassVar[ImplementationRegistry[type, Callable[[Any], Hashable]]] = ImplementationRegistry(issubclass)

    def __new__(
        cls: type[Self]
    ) -> Never:
        raise TypeError

    @decomposers.register(False)
    @staticmethod
    def _[T](
        data: T
    ) -> tuple[T, ...]:
        return (data,)

    @decomposers.register(True)
    @staticmethod
    def _[T](
        data: tuple[T, ...]
    ) -> tuple[T, ...]:
        return data

    @composers.register(False)
    @staticmethod
    def _[T](
        elements: tuple[T, ...]
    ) -> T:
        (element,) = elements
        return element

    @composers.register(True)
    @staticmethod
    def _[T](
        elements: tuple[T, ...]
    ) -> tuple[T, ...]:
        return elements

    @hashers.register(int)
    @hashers.register(str)
    @hashers.register(bytes)
    # Note, for tuples, we require every field should be hashable.
    # Consider creating a data class if it is not the case.
    @hashers.register(tuple)
    @hashers.register(Enum)
    @staticmethod
    def _(
        element: Hashable
    ) -> Hashable:
        return element

    @hashers.register(np.ndarray)
    @staticmethod
    def _(
        element: np.ndarray
    ) -> Hashable:
        return (element.shape, element.dtype, element.tobytes())

    @hashers.register(object)
    @staticmethod
    def _(
        element: object
    ) -> Hashable:
        # We are safe to use `id` here since the memoization is a weak-value-dictionary.
        # The reallocation of an old id indicates that the object has been garbage collected,
        # and its corresponding item in the memoization should have already been removed.
        return id(element)



from __future__ import annotations


from typing import Self

from ...animatables.animatable.piecewiser import Piecewiser
from ...mobjects.mobject import Mobject
from ..timeline import Timeline


class PartialPiecewiser(Piecewiser):
    __slots__ = ()

    def get_segment(
        self: Self,
        alpha: float
    ) -> tuple[float, float]:
        return (0.0, alpha)


class Create(Timeline):
    __slots__ = (
        "_mobject",
        "_piecewiser"
    )

    def __init__(
        self: Self,
        mobject: Mobject,
        *,
        n_segments: int = 1,
        backwards: bool = False
    ) -> None:
        super().__init__(run_alpha=1.0)
        self._mobject: Mobject = mobject
        self._piecewiser: PartialPiecewiser = PartialPiecewiser(
            n_segments=n_segments,
            backwards=backwards
        )

    async def construct(
        self: Self
    ) -> None:
        mobject = self._mobject

        self.scene.add(mobject)
        await self.play(mobject.animate().piecewise(mobject.copy(), self._piecewiser))



from __future__ import annotations


from typing import Self

import numpy as np

from ...constants.constants import TAU
from .polygon import Polygon


class RegularPolygon(Polygon):
    __slots__ = ()

    def __init__(
        self: Self,
        n: int
    ) -> None:
        # By default, one of coordinates is at (1, 0).
        complex_coordinates = np.exp(1.0j * np.linspace(0.0, TAU, n, endpoint=False))
        super().__init__(
            coordinates=np.vstack((complex_coordinates.real, complex_coordinates.imag)).T
        )



from __future__ import annotations


from typing import Self

import moderngl
import numpy as np

from ...constants.custom_typing import (
    NP_xi4,
    ShapeType
)
from ...lazy.lazy import Lazy
from ...toplevel.toplevel import Toplevel
from ..mgl_enums import PrimitiveMode
from ..field import (
    AtomicField,
    Field,
    StructuredField
)
from .buffer import Buffer


class AttributesBuffer(Buffer):
    __slots__ = ()

    def __init__(
        self: Self,
        *,
        field_declarations: tuple[str, ...],
        data_dict: dict[str, np.ndarray],
        index: NP_xi4 | None = None,
        primitive_mode: PrimitiveMode,
        vertices_count: int,
        array_lens: dict[str, int] | None = None
    ) -> None:
        super().__init__(
            shape=(vertices_count,),
            array_lens=array_lens
        )
        self._field_declarations_ = field_declarations
        self._data_dict_ = data_dict
        self._vertices_count_ = vertices_count
        if index is not None:
            self._index_bytes_ = index.astype(np.uint32).tobytes()
            self._use_index_buffer_ = True
        self._primitive_mode_ = primitive_mode

    @Lazy.variable(plural=True)
    @staticmethod
    def _field_declarations_() -> tuple[str, ...]:
        return ()

    @Lazy.variable()
    @staticmethod
    def _data_dict_() -> dict[str, np.ndarray]:
        return {}

    @Lazy.variable()
    @staticmethod
    def _vertices_count_() -> int:
        return 0

    @Lazy.variable()
    @staticmethod
    def _index_bytes_() -> bytes:
        return b""

    @Lazy.variable()
    @staticmethod
    def _use_index_buffer_() -> bool:
        return False

    @Lazy.variable()
    @staticmethod
    def _primitive_mode_() -> PrimitiveMode:
        return NotImplemented

    @Lazy.property(plural=True)
    @staticmethod
    def _fields_(
        field_declarations: tuple[str, ...],
        array_len_items: tuple[tuple[str, int], ...]
    ) -> tuple[AtomicField, ...]:
        return tuple(
            Field.parse_atomic_field(
                field_declaration=field_declaration,
                array_lens_dict=dict(array_len_items)
            )
            for field_declaration in field_declarations
        )

    @Lazy.property()
    @staticmethod
    def _merged_field_(
        fields: tuple[AtomicField, ...]
    ) -> StructuredField:
        return StructuredField(
            name="",
            shape=(),
            fields=fields
        )

    @Lazy.property()
    @staticmethod
    def _buffer_(
        merged_field: StructuredField,
        shape: ShapeType,
        data_dict: dict[str, np.ndarray]
    ) -> moderngl.Buffer:
        return Toplevel._get_context().buffer(data=merged_field.write(shape, data_dict))

    @Lazy.property()
    @staticmethod
    def _index_buffer_(
        index_bytes: bytes,
        use_index_buffer: bool
    ) -> moderngl.Buffer | None:
        if not use_index_buffer:
            return None
        return Toplevel._get_context().buffer(data=index_bytes)



from __future__ import annotations


import itertools
import re
from typing import Self

import attrs
import moderngl

from ..lazy.lazy import Lazy
from ..lazy.lazy_object import LazyObject
from ..toplevel.toplevel import Toplevel
from .buffers.attributes_buffer import AttributesBuffer
from .buffers.texture_buffer import TextureBuffer
from .buffers.uniform_block_buffer import UniformBlockBuffer
from .field import (
    AtomicField,
    StructuredField
)


@attrs.frozen(kw_only=True)
class ProgramUniformInfo:
    array_length: int
    binding: int

    def verify_texture_buffer(
        self: Self,
        texture_buffer: TextureBuffer
    ) -> None:
        assert len(texture_buffer._textures_) == self.array_length


@attrs.frozen(kw_only=True)
class ProgramUniformBlockInfo:
    size: int
    binding: int

    def verify_structured_field(
        self: Self,
        field: StructuredField
    ) -> None:
        assert field._itemsize_ == self.size


@attrs.frozen(kw_only=True)
class ProgramAttributeInfo:
    array_length: int
    dimension: int
    shape: str

    def verify_atomic_field(
        self: Self,
        field: AtomicField
    ) -> None:
        assert field._size_ == self.array_length
        assert field._col_len_ * field._row_len_ == self.dimension
        assert field._base_char_.replace("u", "I") == self.shape


@attrs.frozen(kw_only=True)
class VertexArrayInfo:
    texture_bindings: tuple[tuple[moderngl.Texture, int], ...]
    uniform_block_bindings: tuple[tuple[str, int], ...]
    vertex_array: moderngl.VertexArray


class VertexArray(LazyObject):
    __slots__ = ()

    def __init__(
        self: Self,
        *,
        shader_filename: str,
        custom_macros: tuple[str, ...] = (),
        texture_buffers: tuple[TextureBuffer, ...] = (),
        uniform_block_buffers: tuple[UniformBlockBuffer, ...] = (),
        attributes_buffer: AttributesBuffer
    ) -> None:
        super().__init__()
        self._shader_filename_ = shader_filename
        self._custom_macros_ = custom_macros
        self._texture_buffers_ = texture_buffers
        self._uniform_block_buffers_ = uniform_block_buffers
        self._attributes_buffer_ = attributes_buffer

    @Lazy.variable()
    @staticmethod
    def _shader_filename_() -> str:
        return ""

    @Lazy.variable(plural=True)
    @staticmethod
    def _custom_macros_() -> tuple[str, ...]:
        return ()

    @Lazy.variable(plural=True)
    @staticmethod
    def _texture_buffers_() -> tuple[TextureBuffer, ...]:
        return ()

    @Lazy.variable(plural=True)
    @staticmethod
    def _uniform_block_buffers_() -> tuple[UniformBlockBuffer, ...]:
        return ()

    @Lazy.variable()
    @staticmethod
    def _attributes_buffer_() -> AttributesBuffer:
        return NotImplemented

    @Lazy.property(plural=True)
    @staticmethod
    def _macros_(
        custom_macros: tuple[str, ...],
        texture_buffers__macros: tuple[tuple[str, ...], ...],
        uniform_block_buffers__macros: tuple[tuple[str, ...], ...],
        attributes_buffer__macros: tuple[str, ...]
    ) -> tuple[str, ...]:
        return tuple(itertools.chain(
            custom_macros,
            itertools.chain.from_iterable(texture_buffers__macros),
            itertools.chain.from_iterable(uniform_block_buffers__macros),
            attributes_buffer__macros
        ))

    @Lazy.property()
    @staticmethod
    def _program_(
        shader_filename: str,
        macros: tuple[str, ...],
    ) -> moderngl.Program:

        def read_shader_with_includes_replaced(
            shader_filename: str
        ) -> str:
            for shader_dir in Toplevel._get_config().shader_search_dirs:
                if (shader_path := shader_dir.joinpath(shader_filename)).exists():
                    break
            else:
                raise FileNotFoundError(shader_filename)

            shader_text = shader_path.read_text(encoding="utf-8")
            return re.sub(
                r"#include \"(.+?)\"",
                lambda match: read_shader_with_includes_replaced(match.group(1)),
                shader_text
            )

        shader_text = read_shader_with_includes_replaced(shader_filename)
        shaders = {
            shader_type: "\n".join((
                f"#version {Toplevel._get_context().version_code} core",
                "\n",
                f"#define {shader_type}",
                *macros,
                "\n",
                shader_text
            ))
            for shader_type in (
                "VERTEX_SHADER",
                "FRAGMENT_SHADER",
                "GEOMETRY_SHADER",
                "TESS_CONTROL_SHADER",
                "TESS_EVALUATION_SHADER"
            )
            if re.search(rf"\b{shader_type}\b", shader_text, flags=re.MULTILINE) is not None
        }
        return Toplevel._get_context().program(
            vertex_shader=shaders["VERTEX_SHADER"],
            fragment_shader=shaders.get("FRAGMENT_SHADER"),
            geometry_shader=shaders.get("GEOMETRY_SHADER"),
            tess_control_shader=shaders.get("TESS_CONTROL_SHADER"),
            tess_evaluation_shader=shaders.get("TESS_EVALUATION_SHADER")
        )

    @Lazy.property()
    @staticmethod
    def _vertex_array_info_(
        program: moderngl.Program,
        texture_buffers: tuple[TextureBuffer, ...],
        uniform_block_buffers__field: tuple[StructuredField, ...],
        attributes_buffer: AttributesBuffer
    ) -> VertexArrayInfo | None:
        uniform_info_dict: dict[str, ProgramUniformInfo] = {}
        uniform_block_info_dict: dict[str, ProgramUniformBlockInfo] = {}
        attribute_info_dict: dict[str, ProgramAttributeInfo] = {}
        texture_binding = 1
        uniform_block_binding = 0
        for name in program:
            assert re.fullmatch(r"\w+", name)

            match program[name]:
                case moderngl.Uniform() as uniform:
                    # Used as a `sampler2D`.
                    assert uniform.dimension == 1
                    array_length = uniform.array_length
                    uniform_info_dict[name] = ProgramUniformInfo(
                        array_length=array_length,
                        binding=texture_binding
                    )
                    uniform.value = (
                        texture_binding
                        if array_length == 1
                        else range(texture_binding, texture_binding + array_length)
                    )
                    texture_binding += array_length

                case moderngl.UniformBlock() as uniform_block:
                    uniform_block_info_dict[name] = ProgramUniformBlockInfo(
                        size=uniform_block.size,
                        binding=uniform_block_binding
                    )
                    uniform_block.binding = uniform_block_binding
                    uniform_block_binding += 1

                case moderngl.Attribute() as attribute:
                    attribute_info_dict[name] = ProgramAttributeInfo(
                        array_length=attribute.array_length,
                        dimension=attribute.dimension,
                        shape=attribute.shape
                    )

        texture_bindings: list[tuple[moderngl.Texture, int]] = []
        for texture_buffer in texture_buffers:
            if (uniform_info := uniform_info_dict.pop(texture_buffer._name_, None)) is None:
                continue
            uniform_info.verify_texture_buffer(texture_buffer)
            texture_bindings.extend(
                (texture, binding)
                for binding, texture in enumerate(texture_buffer._textures_, start=uniform_info.binding)
            )

        uniform_block_bindings: list[tuple[str, int]] = []
        for field in uniform_block_buffers__field:
            if (uniform_block_info := uniform_block_info_dict.pop(field._name_, None)) is None:
                continue
            uniform_block_info.verify_structured_field(field)
            uniform_block_bindings.append((field._name_, uniform_block_info.binding))

        attribute_names: list[str] = []
        format_components: list[str] = []
        for field, padding in zip(attributes_buffer._fields_, attributes_buffer._merged_field_._paddings_, strict=True):
            if (attribute_info := attribute_info_dict.pop(field._name_, None)) is None:
                if (total_padding := field._itemsize_ * field._size_ + padding):
                    format_components.append(f"{total_padding}x")
                continue
            attribute_info.verify_atomic_field(field)
            attribute_names.append(field._name_)
            for _ in range(field._size_):
                for _ in range(field._row_len_):
                    format_components.append(f"{field._col_len_}{field._base_char_}{field._base_itemsize_}")
                    if (col_padding := field._col_padding_):
                        format_components.append(f"{col_padding}x{field._base_itemsize_}")
            if padding:
                format_components.append(f"{padding}x")
        format_components.append("/v")

        assert not uniform_info_dict
        assert not uniform_block_info_dict
        assert not attribute_info_dict

        if (
            not attributes_buffer._vertices_count_
            or not attributes_buffer._merged_field_._itemsize_
            or attributes_buffer._use_index_buffer_ and not attributes_buffer._index_bytes_
        ):
            return None
        return VertexArrayInfo(
            texture_bindings=tuple(texture_bindings),
            uniform_block_bindings=tuple(uniform_block_bindings),
            vertex_array=Toplevel._get_context().vertex_array(
                program=program,
                attributes_buffer=attributes_buffer._buffer_,
                attributes_buffer_format_str=" ".join(format_components),
                attribute_names=tuple(attribute_names),
                index_buffer=attributes_buffer._index_buffer_,
                mode=attributes_buffer._primitive_mode_
            )
        )



import json

import sublime
import sublime_plugin


class ExportHighlightCommand(sublime_plugin.ApplicationCommand):
    def run(self):
        window = sublime.active_window()
        view = window.active_view()
        if view.is_loading():
            return

        full_region = sublime.Region(0, view.size())
        tokens = [
            {
                "begin": region.begin(),
                "end": region.end(),
                "style": view.style_for_scope(scope)
            }
            for region, scope in view.extract_tokens_with_scopes(full_region)
        ]
        stem, _ = view.file_name().rsplit(".", maxsplit=1)
        with open(stem + ".json", "w") as output_file:
            json.dump(tokens, output_file)

        view.close()



from __future__ import annotations


import re

from typing import Self

import numpy as np
from colour import Color

from ...constants.custom_typing import (
    ColorType,
    NP_3f8
)
from .animatable_array import AnimatableArray


class AnimatableColor(AnimatableArray[NP_3f8]):
    __slots__ = ()

    def __init__(
        self: Self,
        color: ColorType
    ) -> None:
        super().__init__(type(self)._color_to_array(color))

    @classmethod
    def _color_to_array(
        cls: type[Self],
        color: ColorType
    ) -> NP_3f8:
        match color:
            case Color():
                return np.array(color.rgb)
            case str() if re.fullmatch(r"\w+", color):
                return np.array(Color(color).rgb)
            case str() if re.fullmatch(r"#[0-9A-F]{6}", color, flags=re.IGNORECASE):
                return np.fromiter((
                    int(match.group(), 16)
                    for match in re.finditer(r"[0-9A-F]{2}", color, flags=re.IGNORECASE)
                ), dtype=np.float64) / 255.0
            case np.ndarray() if color.shape == (3,):
                return color.astype(np.float64)
            case _:
                raise ValueError(f"Invalid color: {color}")

    @classmethod
    def _array_to_hex(
        cls: type[Self],
        color_array: NP_3f8
    ) -> str:
        components = (color_array * 255.0).astype(np.int32)
        return f"#{"".join("{:02x}".format(component) for component in components)}"

    @classmethod
    def _color_to_hex(
        cls: type[Self],
        color: ColorType
    ) -> str:
        return cls._array_to_hex(cls._color_to_array(color))



from __future__ import annotations


import itertools
import weakref
from typing import (
    TYPE_CHECKING,
    Callable,
    ClassVar,
    Hashable,
    Iterator,
    Never,
    Self,
    overload
)

if TYPE_CHECKING:
    from .lazy_object import LazyObject
    from .lazy_slot import LazySlot


class Memoized[T]:
    __slots__ = (
        "__weakref__",
        "_id",
        "_value"
    )

    _id_counter: ClassVar[itertools.count[int]] = itertools.count()

    def __init__(
        self: Self,
        value: T
    ) -> None:
        super().__init__()
        self._id: int = next(type(self)._id_counter)
        self._value: T = value

    def get_id(
        self: Self
    ) -> int:
        return self._id

    def get_value(
        self: Self
    ) -> T:
        return self._value


class Memoization[KT: Hashable, VT](weakref.WeakValueDictionary[KT, Memoized[VT]]):
    __slots__ = ()

    def memoize(
        self: Self,
        key: KT,
        value: VT
    ) -> Memoized[VT]:
        if (memoized_value := self.get(key)) is None:
            memoized_value = Memoized(value)
            self[key] = memoized_value
        return memoized_value


class Cache[KT: Hashable, VT](dict[KT, VT]):
    __slots__ = ("_capacity",)

    def __init__(
        self: Self,
        capacity: int
    ) -> None:
        super().__init__()
        self._capacity: int = capacity

    def set(
        self: Self,
        key: KT,
        value: VT
    ) -> None:
        assert key not in self
        self[key] = value
        if len(self) > self._capacity:
            self.pop(next(iter(self)))


type TupleTree[T] = T | tuple[TupleTree[T], ...]


class Tree[T]:
    __slots__ = (
        "_content",
        "_children"
    )

    def __init__(
        self: Self,
        content: T,
        children: tuple[Tree[T], ...] | None = None
    ) -> None:
        super().__init__()
        self._content: T = content
        self._children: tuple[Tree[T], ...] | None = children

    def iter_leaves(
        self: Self
    ) -> Iterator[Tree[T]]:
        if (children := self._children) is not None:
            for child in children:
                yield from child.iter_leaves()
        else:
            yield self

    def as_tuple_tree[ConvertedT](
        self: Self,
        func: Callable[[T], ConvertedT]
    ) -> TupleTree[ConvertedT]:
        if (children := self._children) is not None:
            return tuple(
                child.as_tuple_tree(func)
                for child in children
            )
        return func(self._content)


class LazyDescriptor[T, DataT]:
    __slots__ = (
        "__weakref__",
        "_method",
        "_is_property",
        "_plural",
        "_freeze",
        "_deepcopy",
        "_cache",
        "_parameter_key_memoization",
        "_element_memoization",
        "_name",
        "_parameter_name_chains",
        "_decomposer",
        "_composer",
        "_hasher"
    )

    def __init__(
        self: Self,
        method: Callable[..., DataT],
        is_property: bool,
        plural: bool,
        freeze: bool,
        deepcopy: bool,
        cache_capacity: int
    ) -> None:
        super().__init__()
        self._method: Callable[..., DataT] = method
        self._is_property: bool = is_property
        self._plural: bool = plural
        self._freeze: bool = freeze
        self._deepcopy: bool = deepcopy
        self._cache: Cache[Memoized[Hashable], tuple[Memoized[T], ...]] = Cache(capacity=cache_capacity)
        self._parameter_key_memoization: Memoization[Hashable, Hashable] = Memoization()
        self._element_memoization: Memoization[Hashable, T] = Memoization()
        self._name: str = NotImplemented
        self._parameter_name_chains: tuple[tuple[str, ...], ...] = NotImplemented
        self._decomposer: Callable[[DataT], tuple[T, ...]] = NotImplemented
        self._composer: Callable[[tuple[T, ...]], DataT] = NotImplemented
        self._hasher: Callable[[T], Hashable] = NotImplemented

    @overload
    def __get__(
        self: Self,
        instance: LazyObject,
        owner: type[LazyObject] | None = None
    ) -> DataT: ...

    @overload
    def __get__(
        self: Self,
        instance: None,
        owner: type[LazyObject] | None = None
    ) -> Self: ...

    def __get__(
        self: Self,
        instance: LazyObject | None,
        owner: type[LazyObject] | None = None
    ) -> Self | DataT:
        if instance is None:
            return self
        return self._composer(self.get_elements(instance))

    def __set__(
        self: Self,
        instance: LazyObject,
        data: DataT
    ) -> None:
        self.set_elements(instance, self._decomposer(data))

    def __delete__(
        self: Self,
        instance: LazyObject
    ) -> Never:
        raise TypeError("Cannot delete attributes of a lazy object")

    def _memoize_parameter_key(
        self: Self,
        parameter_key: Hashable
    ) -> Memoized[Hashable]:
        return self._parameter_key_memoization.memoize(parameter_key, parameter_key)

    def _memoize_elements(
        self: Self,
        elements: tuple[T, ...]
    ) -> tuple[Memoized[T], ...]:
        element_memoization = self._element_memoization
        hasher = self._hasher
        return tuple(
            element_memoization.memoize(hasher(element), element)
            for element in elements
        )

    def _get_memoized_elements(
        self: Self,
        instance: LazyObject
    ) -> tuple[Memoized[T], ...]:
        slot = self.get_slot(instance)
        if (memoized_elements := slot.get()) is None:
            # If there's at least a parameter, `slot` is guaranteed to be a property slot.
            # Associate it with variable slots.
            tree_root = Memoized(instance)
            trees = tuple(Tree(tree_root) for _ in self._parameter_name_chains)
            associated_slots: set[LazySlot] = set()
            for parameter_name_chain, tree in zip(self._parameter_name_chains, trees, strict=True):
                for name in parameter_name_chain:
                    for leaf in tree.iter_leaves():
                        leaf_object = leaf._content.get_value()
                        leaf_slot = leaf_object._get_lazy_slot(name)
                        descriptor = leaf_slot.get_descriptor()
                        children_memoized_elements = descriptor._get_memoized_elements(leaf_object)
                        if descriptor._plural:
                            leaf._children = tuple(Tree(element) for element in children_memoized_elements)
                        else:
                            (leaf._content,) = children_memoized_elements
                        if descriptor._is_property:
                            associated_slots.update(leaf_slot.iter_associated_slots())
                        else:
                            associated_slots.add(leaf_slot)

            memoized_parameter_key = self._memoize_parameter_key(tuple(
                tree.as_tuple_tree(Memoized.get_id) for tree in trees
            ))
            if (memoized_elements := self._cache.get(memoized_parameter_key)) is None:
                memoized_elements = self._memoize_elements(self._decomposer(self._method(*(
                    tree.as_tuple_tree(Memoized.get_value) for tree in trees
                ))))
                self._cache.set(memoized_parameter_key, memoized_elements)
            slot.set(
                elements=memoized_elements,
                parameter_key=memoized_parameter_key,
                associated_slots=associated_slots
            )
        return memoized_elements

    def _set_memoized_elements(
        self: Self,
        instance: LazyObject,
        memoized_elements: tuple[Memoized[T], ...]
    ) -> None:
        assert not self._is_property
        slot = self.get_slot(instance)
        # Guaranteed to be a variable slot. Expire associated property slots.
        for expired_property_slot in slot.iter_associated_slots():
            expired_property_slot.expire()
        slot.set(
            elements=memoized_elements,
            parameter_key=None,
            associated_slots=set()
        )

    def get_slot(
        self: Self,
        instance: LazyObject
    ) -> LazySlot:
        return instance._get_lazy_slot(self._name)

    def get_elements(
        self: Self,
        instance: LazyObject
    ) -> tuple[T, ...]:
        return tuple(memoized_element.get_value() for memoized_element in self._get_memoized_elements(instance))

    def set_elements(
        self: Self,
        instance: LazyObject,
        elements: tuple[T, ...]
    ) -> None:
        self._set_memoized_elements(instance, self._memoize_elements(elements))



from __future__ import annotations


import asyncio
from abc import (
    ABC,
    abstractmethod
)
from typing import (
    TYPE_CHECKING,
    Callable,
    Coroutine,
    Self
)

import attrs

from ..constants.custom_typing import (
    ConditionType,
    RateType
)
from ..constants.rates import Rates
from ..toplevel.toplevel import Toplevel

if TYPE_CHECKING:
    from ..toplevel.scene import Scene


@attrs.frozen(kw_only=True)
class TimelineState:
    pass


@attrs.frozen(kw_only=True)
class InitialState(TimelineState):
    pass


@attrs.frozen(kw_only=True)
class AfterScheduled(TimelineState):
    parent_absolute_rate: Callable[[], float]
    rate: RateType
    run_time_scale: float
    run_alpha_scale: float


@attrs.frozen(kw_only=True)
class AfterLaunched(TimelineState):
    construct_coroutine: Coroutine[None, None, None]
    absolute_rate: Callable[[], float]
    children: list[Timeline]


@attrs.frozen(kw_only=True)
class AfterTerminated(TimelineState):
    pass


class Timeline(ABC):
    __slots__ = (
        "__weakref__",
        "_run_alpha",
        "_timeline_state",
        "_launch_condition",
        "_terminate_condition",
        "_progress_condition"
    )

    def __init__(
        self: Self,
        # The accumulated alpha value of `construct`.
        # Left as `inf` if infinite or indefinite.
        # This parameter is required mostly for the program to know
        # how long the timeline is before running the timeline.
        run_alpha: float = float("inf")
    ) -> None:
        super().__init__()
        self._run_alpha: float = run_alpha
        self._timeline_state: TimelineState = InitialState()
        self._launch_condition: ConditionType | None = None  # default: lambda: True
        self._terminate_condition: ConditionType | None = None  # default: lambda: False
        self._progress_condition: ConditionType | None = None  # default: lambda: True

    def _progress(
        self: Self
    ) -> None:
        assert self.scheduled()
        if isinstance(self._timeline_state, AfterScheduled):
            if self._launch_condition is not None and not self._launch_condition():
                return
            self.launch()
        if isinstance(self._timeline_state, AfterLaunched):
            self._animation_update(self._timeline_state.absolute_rate())
            while self._terminate_condition is None or not self._terminate_condition():
                for child in self._timeline_state.children[:]:
                    child._progress()
                    if child.terminated():
                        self._timeline_state.children.remove(child)
                if self._progress_condition is not None and not self._progress_condition():
                    return
                try:
                    self._timeline_state.construct_coroutine.send(None)
                except StopIteration:
                    break
            self.terminate()
        if isinstance(self._timeline_state, AfterTerminated):
            return

    def _animation_update(
        self: Self,
        time: float
    ) -> None:
        pass

    @abstractmethod
    async def construct(
        self: Self
    ) -> None:
        pass

    def scheduled(
        self: Self
    ) -> bool:
        return isinstance(self._timeline_state, AfterScheduled | AfterLaunched | AfterTerminated)

    def launched(
        self: Self
    ) -> bool:
        return isinstance(self._timeline_state, AfterLaunched | AfterTerminated)

    def terminated(
        self: Self
    ) -> bool:
        return isinstance(self._timeline_state, AfterTerminated)

    def schedule(
        self: Self,
        parent_absolute_rate: Callable[[], float],
        rate: RateType = Rates.linear(),
        run_time_scale: float = 1.0,
        run_alpha_scale: float = 1.0,
        launch_condition: ConditionType | None = None,
        terminate_condition: ConditionType | None = None
    ) -> None:
        assert isinstance(self._timeline_state, InitialState)
        self._timeline_state = AfterScheduled(
            parent_absolute_rate=parent_absolute_rate,
            rate=rate,
            run_time_scale=run_time_scale,
            run_alpha_scale=run_alpha_scale
        )
        if launch_condition is not None:
            self._launch_condition = launch_condition
        if terminate_condition is not None:
            self._terminate_condition = terminate_condition

    def launch(
        self: Self,
        initial_progress_condition: ConditionType | None = None
    ) -> None:
        assert isinstance(self._timeline_state, AfterScheduled)
        parent_absolute_rate = self._timeline_state.parent_absolute_rate
        rate = self._timeline_state.rate
        run_time_scale = self._timeline_state.run_time_scale
        run_alpha_scale = self._timeline_state.run_alpha_scale
        initial_alpha = parent_absolute_rate()
        self._timeline_state = AfterLaunched(
            construct_coroutine=self.construct(),
            absolute_rate=lambda: rate((parent_absolute_rate() - initial_alpha) / run_time_scale) * run_alpha_scale,
            children=[]
        )
        if initial_progress_condition is not None:
            self._progress_condition = initial_progress_condition

    def terminate(
        self: Self
    ) -> None:
        assert isinstance(self._timeline_state, AfterLaunched)
        self._timeline_state = AfterTerminated()
        self._launch_condition = None
        self._terminate_condition = None
        self._progress_condition = None

    def prepare(
        self: Self,
        timeline: Timeline,
        *,
        # `[0.0, +infty) -> [0.0, +infty), time |-> alpha`
        # Must be an increasing function.
        # Must be `None` if `_run_alpha` is infinity.
        rate: RateType | None = None,
        # Intepreted as "the inverse of run speed" if `_run_alpha` is infinity.
        run_time: float | None = None,
        launch_condition: ConditionType | None = None,
        terminate_condition: ConditionType | None = None
    ) -> None:
        assert isinstance(self._timeline_state, AfterLaunched)
        if (run_alpha := timeline._run_alpha) == float("inf"):
            assert rate is None
            run_alpha_scale = 1.0
        else:
            run_alpha_scale = run_alpha
        if run_time is None:
            run_time_scale = run_alpha_scale
        else:
            run_time_scale = run_time
        if rate is None:
            rate = Rates.linear()
        timeline.schedule(
            parent_absolute_rate=self._timeline_state.absolute_rate,
            rate=rate,
            run_time_scale=run_time_scale,
            run_alpha_scale=run_alpha_scale,
            launch_condition=launch_condition,
            terminate_condition=terminate_condition
        )
        self._timeline_state.children.append(timeline)

    async def wait_until(
        self: Self,
        progress_condition: ConditionType
    ) -> None:
        self._progress_condition = progress_condition
        await asyncio.sleep(0.0)

    # shortcuts

    @property
    def scene(
        self: Self
    ) -> Scene:
        return Toplevel._get_scene()

    async def play(
        self: Self,
        timeline: Timeline,
        rate: RateType | None = None,
        run_time: float | None = None,
        launch_condition: ConditionType | None = None,
        terminate_condition: ConditionType | None = None
    ) -> None:
        self.prepare(
            timeline,
            rate=rate,
            run_time=run_time,
            launch_condition=launch_condition,
            terminate_condition=terminate_condition
        )
        await self.wait_until(timeline.terminated)

    async def wait(
        self: Self,
        delta_alpha: float = 1.0
    ) -> None:
        assert isinstance(self._timeline_state, AfterLaunched)
        target_alpha = self._timeline_state.absolute_rate() + delta_alpha
        await self.wait_until(
            lambda: not isinstance(self._timeline_state, AfterLaunched) or self._timeline_state.absolute_rate() >= target_alpha
        )



from __future__ import annotations


from typing import (
    Iterator,
    Self
)

import moderngl

from ...animatables.animatable.animatable import AnimatableActions
from ...animatables.arrays.animatable_color import AnimatableColor
from ...animatables.arrays.animatable_float import AnimatableFloat
from ...animatables.mesh import Mesh
from ...animatables.model import ModelActions
from ...constants.custom_typing import (
    NP_3f8,
    NP_f8,
    NP_x2f8,
    NP_x3f8,
    NP_x3i4
)
from ...lazy.lazy import Lazy
from ...rendering.buffers.attributes_buffer import AttributesBuffer
from ...rendering.buffers.texture_buffer import TextureBuffer
from ...rendering.buffers.uniform_block_buffer import UniformBlockBuffer
from ...rendering.mgl_enums import PrimitiveMode
from ...rendering.vertex_array import VertexArray
from ...toplevel.toplevel import Toplevel
from ..mobject import Mobject


class MeshMobject(Mobject):
    __slots__ = ()

    def __init__(
        self: Self,
        mesh: Mesh | None = None
    ) -> None:
        super().__init__()
        if mesh is not None:
            self._mesh_ = mesh

    @AnimatableActions.interpolate.register_descriptor()
    @Lazy.volatile()
    @staticmethod
    def _mesh_() -> Mesh:
        return Mesh()

    @AnimatableActions.interpolate.register_descriptor()
    @ModelActions.set.register_descriptor(converter=AnimatableColor)
    @Lazy.volatile()
    @staticmethod
    def _color_() -> AnimatableColor:
        return AnimatableColor(Toplevel._get_config().default_color)

    @AnimatableActions.interpolate.register_descriptor()
    @ModelActions.set.register_descriptor(converter=AnimatableFloat)
    @Lazy.volatile()
    @staticmethod
    def _opacity_() -> AnimatableFloat:
        return AnimatableFloat(Toplevel._get_config().default_opacity)

    @AnimatableActions.interpolate.register_descriptor()
    @ModelActions.set.register_descriptor(converter=AnimatableFloat)
    @Lazy.volatile()
    @staticmethod
    def _weight_() -> AnimatableFloat:
        return AnimatableFloat(Toplevel._get_config().default_weight)

    @AnimatableActions.interpolate.register_descriptor()
    @ModelActions.set.register_descriptor(converter=AnimatableFloat)
    @Lazy.volatile()
    @staticmethod
    def _ambient_strength_() -> AnimatableFloat:
        return AnimatableFloat(Toplevel._get_config().mesh_ambient_strength)

    @AnimatableActions.interpolate.register_descriptor()
    @ModelActions.set.register_descriptor(converter=AnimatableFloat)
    @Lazy.volatile()
    @staticmethod
    def _specular_strength_() -> AnimatableFloat:
        return AnimatableFloat(Toplevel._get_config().mesh_specular_strength)

    @AnimatableActions.interpolate.register_descriptor()
    @ModelActions.set.register_descriptor(converter=AnimatableFloat)
    @Lazy.volatile()
    @staticmethod
    def _shininess_() -> AnimatableFloat:
        return AnimatableFloat(Toplevel._get_config().mesh_shininess)

    @Lazy.variable(plural=True)
    @staticmethod
    def _color_maps_() -> tuple[moderngl.Texture, ...]:
        return ()

    @Lazy.property()
    @staticmethod
    def _local_sample_positions_(
        mesh__positions: NP_x3f8,
        mesh__faces: NP_x3i4
    ) -> NP_x3f8:
        return mesh__positions[mesh__faces.flatten()]

    @Lazy.property()
    @staticmethod
    def _color_maps_texture_buffer_(
        color_maps: tuple[moderngl.Texture, ...]
    ) -> TextureBuffer:
        return TextureBuffer(
            name="t_color_maps",
            textures=color_maps,
            array_lens={
                "NUM_T_COLOR_MAPS": len(color_maps)
            }
        )

    @Lazy.property()
    @staticmethod
    def _material_uniform_block_buffer_(
        color__array: NP_3f8,
        opacity__array: NP_f8,
        weight__array: NP_f8,
        ambient_strength__array: NP_f8,
        specular_strength__array: NP_f8,
        shininess__array: NP_f8
    ) -> UniformBlockBuffer:
        return UniformBlockBuffer(
            name="ub_material",
            field_declarations=(
                "vec3 u_color",
                "float u_opacity",
                "float u_weight",
                "float u_ambient_strength",
                "float u_specular_strength",
                "float u_shininess"
            ),
            data_dict={
                "u_color": color__array,
                "u_opacity": opacity__array,
                "u_weight": weight__array,
                "u_ambient_strength": ambient_strength__array,
                "u_specular_strength": specular_strength__array,
                "u_shininess": shininess__array
            }
        )

    @Lazy.property()
    @staticmethod
    def _mesh_attributes_buffer_(
        mesh__positions: NP_x3f8,
        mesh__normals: NP_x3f8,
        mesh__uvs: NP_x2f8,
        mesh__faces: NP_x3i4
    ) -> AttributesBuffer:
        return AttributesBuffer(
            field_declarations=(
                "vec3 in_position",
                "vec3 in_normal",
                "vec2 in_uv"
            ),
            data_dict={
                "in_position": mesh__positions,
                "in_normal": mesh__normals,
                "in_uv": mesh__uvs
            },
            index=mesh__faces.flatten(),
            primitive_mode=PrimitiveMode.TRIANGLES,
            vertices_count=len(mesh__positions)
        )

    @Lazy.property()
    @staticmethod
    def _mesh_vertex_array_(
        color_maps_texture_buffer: TextureBuffer,
        camera__camera_uniform_block_buffer: UniformBlockBuffer,
        lighting__lighting_uniform_block_buffer: UniformBlockBuffer,
        model_uniform_block_buffer: UniformBlockBuffer,
        material_uniform_block_buffer: UniformBlockBuffer,
        mesh_attributes_buffer: AttributesBuffer
    ) -> VertexArray:
        return VertexArray(
            shader_filename="mesh.glsl",
            texture_buffers=(
                color_maps_texture_buffer,
            ),
            uniform_block_buffers=(
                camera__camera_uniform_block_buffer,
                lighting__lighting_uniform_block_buffer,
                model_uniform_block_buffer,
                material_uniform_block_buffer
            ),
            attributes_buffer=mesh_attributes_buffer
        )

    def _iter_vertex_arrays(
        self: Self
    ) -> Iterator[VertexArray]:
        yield from super()._iter_vertex_arrays()
        yield self._mesh_vertex_array_



from __future__ import annotations


from ...constants.constants import ORIGIN
from ...constants.custom_typing import (
    NP_3f8,
    NP_44f8
)
from ...lazy.lazy import Lazy
from ...toplevel.toplevel import Toplevel
from ..animatable.animatable import AnimatableActions
from ..arrays.animatable_color import AnimatableColor
from ..arrays.model_matrix import ModelMatrix
from ..model import ModelActions
from ..point import Point


class PointLight(Point):
    __slots__ = ()

    @AnimatableActions.interpolate.register_descriptor()
    @ModelActions.set.register_descriptor(converter=AnimatableColor)
    @Lazy.volatile()
    @staticmethod
    def _color_() -> AnimatableColor:
        return AnimatableColor(Toplevel._get_config().default_color)

    @Lazy.property()
    @staticmethod
    def _position_(
        model_matrix__array: NP_44f8
    ) -> NP_3f8:
        return ModelMatrix._apply(model_matrix__array, ORIGIN)



from __future__ import annotations


import functools
import re
from typing import (
    ClassVar,
    Self
)

import numpy as np

from ..constants.custom_typing import ShapeType
from ..lazy.lazy import Lazy
from ..lazy.lazy_object import LazyObject


class Field(LazyObject):
    __slots__ = ()

    # Note, numpy matrices are row-major, while OpenGL store matrices in column-major format.
    # In numpy, shape `(r, c)` stands for `r` rows and `c` columns.
    # In glsl, `matcxr` specifies a matrix with `r` rows and `c` columns.
    # The behavior of a `matcxr` is simulated by an ndarray with shape `(r, c)`,
    # and one needs to transpose matrices before passing them to shaders.
    _GLSL_DTYPES: ClassVar[dict[str, np.dtype]] = {
        "int":     np.dtype("i4"),
        "ivec2":   np.dtype("2i4"),
        "ivec3":   np.dtype("3i4"),
        "ivec4":   np.dtype("4i4"),
        "uint":    np.dtype("u4"),
        "uvec2":   np.dtype("2u4"),
        "uvec3":   np.dtype("3u4"),
        "uvec4":   np.dtype("4u4"),
        "float":   np.dtype("f4"),
        "vec2":    np.dtype("2f4"),
        "vec3":    np.dtype("3f4"),
        "vec4":    np.dtype("4f4"),
        "mat2":    np.dtype("(2,2)f4"),
        "mat2x3":  np.dtype("(2,3)f4"),
        "mat2x4":  np.dtype("(2,4)f4"),
        "mat3x2":  np.dtype("(3,2)f4"),
        "mat3":    np.dtype("(3,3)f4"),
        "mat3x4":  np.dtype("(3,4)f4"),
        "mat4x2":  np.dtype("(4,2)f4"),
        "mat4x3":  np.dtype("(4,3)f4"),
        "mat4":    np.dtype("(4,4)f4"),
        "double":  np.dtype("f8"),
        "dvec2":   np.dtype("2f8"),
        "dvec3":   np.dtype("3f8"),
        "dvec4":   np.dtype("4f8"),
        "dmat2":   np.dtype("(2,2)f8"),
        "dmat2x3": np.dtype("(2,3)f8"),
        "dmat2x4": np.dtype("(2,4)f8"),
        "dmat3x2": np.dtype("(3,2)f8"),
        "dmat3":   np.dtype("(3,3)f8"),
        "dmat3x4": np.dtype("(3,4)f8"),
        "dmat4x2": np.dtype("(4,2)f8"),
        "dmat4x3": np.dtype("(4,3)f8"),
        "dmat4":   np.dtype("(4,4)f8")
    }

    def __init__(
        self: Self,
        *,
        name: str,
        shape: ShapeType,
        itemsize: int,
        base_alignment: int
    ) -> None:
        super().__init__()
        self._name_ = name
        self._shape_ = shape
        self._itemsize_ = itemsize
        self._base_alignment_ = base_alignment

    @Lazy.variable()
    @staticmethod
    def _name_() -> str:
        return ""

    @Lazy.variable()
    @staticmethod
    def _shape_() -> ShapeType:
        return ()

    @Lazy.variable()
    @staticmethod
    def _itemsize_() -> int:
        return 0

    @Lazy.variable()
    @staticmethod
    def _base_alignment_() -> int:
        return 1

    @Lazy.property()
    @staticmethod
    def _size_(
        shape: ShapeType
    ) -> int:
        return functools.reduce(int.__mul__, shape, 1)

    @Lazy.property()
    @staticmethod
    def _dtype_() -> np.dtype:
        return np.dtype("f4")

    @Lazy.property(plural=True)
    @staticmethod
    def _pointers_() -> tuple[tuple[tuple[str, ...], int], ...]:
        return ()

    @classmethod
    def _parse_field_declaration(
        cls: type[Self],
        field_declaration: str,
        array_lens_dict: dict[str, int]
    ) -> tuple[str, str, ShapeType]:
        pattern = re.compile(r"""
            (?P<dtype_str>\w+?)
            \s
            (?P<name>\w+?)
            (?P<shape>(\[\w+?\])*)
        """, flags=re.VERBOSE)
        match = pattern.fullmatch(field_declaration)
        assert match is not None

        dtype_str = match.group("dtype_str")
        name = match.group("name")
        shape = tuple(
            int(s) if re.match(r"^\d+$", s := index_match.group(1)) is not None else array_lens_dict[s]
            for index_match in re.finditer(r"\[(\w+?)\]", match.group("shape"))
        )
        return dtype_str, name, shape

    @classmethod
    def parse_atomic_field(
        cls: type[Self],
        field_declaration: str,
        array_lens_dict: dict[str, int]
    ) -> AtomicField:
        dtype_str, name, shape = cls._parse_field_declaration(field_declaration, array_lens_dict)
        return AtomicField(
            name=name,
            shape=shape,
            dtype=cls._GLSL_DTYPES[dtype_str]
        )

    @classmethod
    def parse_field(
        cls: type[Self],
        field_declaration: str,
        struct_dict: dict[str, tuple[str, ...]],
        array_lens_dict: dict[str, int]
    ) -> Field:
        dtype_str, name, shape = cls._parse_field_declaration(field_declaration, array_lens_dict)
        if (dtype := cls._GLSL_DTYPES.get(dtype_str)) is not None:
            return AtomicField(
                name=name,
                shape=shape,
                dtype=dtype
            )
        return StructuredField(
            name=name,
            shape=shape,
            fields=tuple(
                cls.parse_field(
                    field_declaration=child_field_declaration,
                    struct_dict=struct_dict,
                    array_lens_dict=array_lens_dict
                )
                for child_field_declaration in struct_dict[dtype_str]
            )
        )

    def _get_np_buffer_and_pointers(
        self: Self,
        shape: ShapeType
    ) -> tuple[np.ndarray, dict[str, tuple[np.ndarray, int]]]:

        def get_np_buffer_pointer(
            np_buffer: np.ndarray,
            name_chain: tuple[str, ...]
        ) -> np.ndarray:
            result = np_buffer
            for name in name_chain:
                result = result[name]
            return result["_"]

        np_buffer = np.zeros(shape, dtype=self._dtype_)
        np_buffer_pointers = {
            ".".join(name_chain): (get_np_buffer_pointer(np_buffer, name_chain), base_ndim)
            for name_chain, base_ndim in self._pointers_
        }
        return np_buffer, np_buffer_pointers

    def write(
        self: Self,
        shape: ShapeType,
        data_dict: dict[str, np.ndarray]
    ) -> bytes:
        np_buffer, np_buffer_pointers = self._get_np_buffer_and_pointers(shape)
        for key, (np_buffer_pointer, base_ndim) in np_buffer_pointers.items():
            data = data_dict[key]
            if not np_buffer_pointer.size:
                assert not data.size
                continue
            data_expanded = np.expand_dims(data, axis=tuple(range(-2, -base_ndim)))
            assert np_buffer_pointer.shape == data_expanded.shape
            np_buffer_pointer[...] = data_expanded
        return np_buffer.tobytes()

    def read(
        self: Self,
        shape: ShapeType,
        data_bytes: bytes
    ) -> dict[str, np.ndarray]:
        data_dict: dict[str, np.ndarray] = {}
        np_buffer, np_buffer_pointers = self._get_np_buffer_and_pointers(shape)
        np_buffer[...] = np.frombuffer(data_bytes, dtype=np_buffer.dtype).reshape(np_buffer.shape)
        for key, (np_buffer_pointer, base_ndim) in np_buffer_pointers.items():
            data_expanded = np_buffer_pointer[...]
            data = np.squeeze(data_expanded, axis=tuple(range(-2, -base_ndim)))
            data_dict[key] = data
        return data_dict


class AtomicField(Field):
    __slots__ = ()

    def __init__(
        self: Self,
        name: str,
        shape: ShapeType,
        dtype: np.dtype
    ) -> None:
        base_shape = dtype.shape
        base_char = dtype.base.char
        base_itemsize = dtype.base.itemsize
        assert len(base_shape) <= 2 and all(2 <= l <= 4 for l in base_shape)
        col_len, row_len, *_ = base_shape + (1, 1)
        col_padding = 0 if not shape and row_len == 1 else 4 - col_len
        base_alignment = (col_len if not shape and col_len <= 2 and row_len == 1 else 4) * base_itemsize

        super().__init__(
            name=name,
            shape=shape,
            itemsize=row_len * (col_len + col_padding) * base_itemsize,
            base_alignment=base_alignment
        )
        self._base_char_ = base_char
        self._base_itemsize_ = base_itemsize
        self._base_ndim_ = len(base_shape)
        self._row_len_ = row_len
        self._col_len_ = col_len
        self._col_padding_ = col_padding

    @Lazy.variable()
    @staticmethod
    def _base_char_() -> str:
        return ""

    @Lazy.variable()
    @staticmethod
    def _base_itemsize_() -> int:
        return 0

    @Lazy.variable()
    @staticmethod
    def _base_ndim_() -> int:
        return 0

    @Lazy.variable()
    @staticmethod
    def _row_len_() -> int:
        return 0

    @Lazy.variable()
    @staticmethod
    def _col_len_() -> int:
        return 0

    @Lazy.variable()
    @staticmethod
    def _col_padding_() -> int:
        return 0

    @Lazy.property()
    @staticmethod
    def _dtype_(
        base_char: str,
        base_itemsize: int,
        col_len: int,
        col_padding: int,
        row_len: int
    ) -> np.dtype:
        return np.dtype((np.dtype({
            "names": ["_"],
            "formats": [(np.dtype(f"{base_char}{base_itemsize}"), (col_len,))],
            "itemsize": (col_len + col_padding) * base_itemsize
        }), (row_len,)))

    @Lazy.property(plural=True)
    @staticmethod
    def _pointers_(
        base_ndim: int
    ) -> tuple[tuple[tuple[str, ...], int], ...]:
        return (((), base_ndim),)


class StructuredField(Field):
    __slots__ = ()

    def __init__(
        self: Self,
        name: str,
        shape: ShapeType,
        fields: tuple[Field, ...]
    ) -> None:
        struct_base_alignment = 16
        next_base_alignments = (*(field._base_alignment_ for field in fields), struct_base_alignment)[1:]
        offsets: list[int] = []
        paddings: list[int] = []
        offset = 0
        for field, next_base_alignment in zip(fields, next_base_alignments, strict=True):
            offsets.append(offset)
            offset += field._itemsize_ * field._size_
            padding = (-offset) % next_base_alignment
            paddings.append(padding)
            offset += padding

        super().__init__(
            name=name,
            shape=shape,
            itemsize=offset,
            base_alignment=struct_base_alignment
        )
        self._fields_ = fields
        self._offsets_ = tuple(offsets)
        self._paddings_ = tuple(paddings)

    @Lazy.variable(plural=True)
    @staticmethod
    def _fields_() -> tuple[Field, ...]:
        return ()

    @Lazy.variable(plural=True)
    @staticmethod
    def _offsets_() -> tuple[int, ...]:
        return ()

    @Lazy.variable(plural=True)
    @staticmethod
    def _paddings_() -> tuple[int, ...]:
        return ()

    @Lazy.property()
    @staticmethod
    def _dtype_(
        fields: tuple[Field, ...],
        offsets: tuple[int, ...],
        itemsize: int
    ) -> np.dtype:
        return np.dtype({
            "names": tuple(field._name_ for field in fields),
            "formats": tuple((field._dtype_, field._shape_) for field in fields),
            "offsets": offsets,
            "itemsize": itemsize
        })

    @Lazy.property(plural=True)
    @staticmethod
    def _pointers_(
        fields: tuple[Field, ...]
    ) -> tuple[tuple[tuple[str, ...], int], ...]:
        return tuple(
            ((field._name_,) + name_chain, base_ndim)
            for field in fields
            for name_chain, base_ndim in field._pointers_
        )



from __future__ import annotations


from typing import Self

import numpy as np

from ...constants.custom_typing import NP_3f8
from .polyline import Polyline


class Line(Polyline):
    __slots__ = ()

    def __init__(
        self: Self,
        position_0: NP_3f8,
        position_1: NP_3f8
    ) -> None:
        super().__init__(
            positions=np.array((position_0, position_1))
        )



from __future__ import annotations


from abc import abstractmethod
from typing import (
    Iterator,
    Self,
    Unpack
)

from ...constants.custom_typing import NP_xf8
from ...lazy.lazy import Lazy
from ...lazy.lazy_object import LazyObject
from .actions import (
    Action,
    Actions,
    DescriptiveAction,
    DescriptorParameters
)
from .animation import (
    AnimateKwargs,
    Animation,
    AnimationsTimeline
)
from .piecewiser import Piecewiser


class AnimatableActions(Actions):
    __slots__ = ()

    @DescriptiveAction.register(DescriptorParameters)
    @classmethod
    def interpolate(
        cls: type[Self],
        dst: Animatable,
        src_0: Animatable,
        src_1: Animatable
    ) -> Iterator[Animation]:
        for descriptor, _ in cls.interpolate.iter_descriptor_items():
            if not all(
                descriptor in animatable._lazy_descriptors
                for animatable in (dst, src_0, src_1)
            ):
                continue
            for dst_element, src_0_element, src_1_element in zip(
                descriptor.get_elements(dst),
                descriptor.get_elements(src_0),
                descriptor.get_elements(src_1),
                strict=True
            ):
                assert isinstance(dst_element, Animatable)
                yield from type(dst_element).interpolate.iter_animations(
                    dst=dst_element,
                    src_0=src_0_element,
                    src_1=src_1_element
                )

    @DescriptiveAction.register(DescriptorParameters)
    @classmethod
    def piecewise(
        cls: type[Self],
        dst: Animatable,
        src: Animatable,
        piecewiser: Piecewiser
    ) -> Iterator[Animation]:
        for descriptor, _ in cls.piecewise.iter_descriptor_items():
            if not all(
                descriptor in animatable._lazy_descriptors
                for animatable in (dst, src)
            ):
                continue
            for dst_element, src_element in zip(
                descriptor.get_elements(dst),
                descriptor.get_elements(src),
                strict=True
            ):
                assert isinstance(dst_element, Animatable)
                yield from type(dst_element).piecewise.iter_animations(
                    dst=dst_element,
                    src=src_element,
                    piecewiser=piecewiser
                )

    @Action.register()
    @classmethod
    def transform(
        cls: type[Self],
        dst: Animatable,
        src: Animatable
    ) -> Iterator[Animation]:
        yield from cls.interpolate.iter_animations(
            dst=dst,
            src_0=dst.copy(),
            src_1=src
        )


class Animatable(AnimatableActions, LazyObject):
    __slots__ = ()

    def animate(
        self: Self,
        **kwargs: Unpack[AnimateKwargs]
    ) -> DynamicAnimatable[Self]:
        return DynamicAnimatable(self, **kwargs)


class DynamicAnimatable[AnimatableT: Animatable](AnimatableActions, AnimationsTimeline):
    __slots__ = ("_dst",)

    def __init__(
        self: Self,
        dst: AnimatableT,
        **kwargs: Unpack[AnimateKwargs]
    ) -> None:
        super().__init__(**kwargs)
        self._dst: AnimatableT = dst


class AnimatableInterpolateAnimation[AnimatableT: Animatable](Animation):
    __slots__ = (
        "_dst",
        "_interpolate_info"
    )

    def __init__(
        self: Self,
        dst: AnimatableT,
        src_0: AnimatableT,
        src_1: AnimatableT
    ) -> None:
        super().__init__()
        self._dst: AnimatableT = dst
        self._src_0_ = src_0.copy()
        self._src_1_ = src_1.copy()

    @Lazy.variable()
    @staticmethod
    def _src_0_() -> AnimatableT:
        return NotImplemented

    @Lazy.variable()
    @staticmethod
    def _src_1_() -> AnimatableT:
        return NotImplemented

    @abstractmethod
    def interpolate(
        self: Self,
        dst: AnimatableT,
        alpha: float
    ) -> None:
        pass

    def update(
        self: Self,
        alpha: float
    ) -> None:
        super().update(alpha)
        self.interpolate(self._dst, alpha)


class AnimatablePiecewiseAnimation[AnimatableT: Animatable](Animation):
    __slots__ = (
        "_dst",
        "_src",
        "_piecewiser"
    )

    def __init__(
        self: Self,
        dst: AnimatableT,
        src: AnimatableT,
        piecewise_func: Piecewiser
    ) -> None:
        super().__init__()
        self._dst: AnimatableT = dst
        self._src: AnimatableT = src
        self._piecewiser: Piecewiser = piecewise_func

    @classmethod
    @abstractmethod
    def split(
        cls: type[Self],
        dsts: tuple[AnimatableT, ...],
        src: AnimatableT,
        alphas: NP_xf8
    ) -> None:
        pass

    @classmethod
    @abstractmethod
    def concatenate(
        cls: type[Self],
        dst: AnimatableT,
        srcs: tuple[AnimatableT, ...]
    ) -> None:
        pass

    def update(
        self: Self,
        alpha: float
    ) -> None:
        super().update(alpha)
        cls = type(self)
        piecewise_info = self._piecewiser.piecewise(alpha)
        dst = self._dst
        animatable_cls = type(dst)
        pieces = tuple(animatable_cls() for _ in range(len(piecewise_info.split_alphas) + 1))
        cls.split(pieces, self._src, piecewise_info.split_alphas)
        cls.concatenate(dst, tuple(pieces[index] for index in piecewise_info.concatenate_indices))



from __future__ import annotations


from typing import Self

import attrs

from .toplevel import Toplevel


@attrs.frozen(kw_only=True)
class EventInfo:
    event_type: str
    symbol: int | None
    modifiers: int | None
    buttons: int | None
    x: int | None
    y: int | None
    dx: int | None
    dy: int | None
    scroll_x: float | None
    scroll_y: float | None


class Event:
    __slots__ = (
        "_event_info",
        "_captured_event_info"
    )

    def __init__(
        self: Self,
        *,
        event_type: str,
        symbol: int | None = None,
        modifiers: int | None = None,
        buttons: int | None = None,
        x: int | None = None,
        y: int | None = None,
        dx: int | None = None,
        dy: int | None = None,
        scroll_x: float | None = None,
        scroll_y: float | None = None
    ) -> None:
        super().__init__()
        self._event_info: EventInfo = EventInfo(
            event_type=event_type,
            symbol=symbol,
            modifiers=modifiers,
            buttons=buttons,
            x=x,
            y=y,
            dx=dx,
            dy=dy,
            scroll_x=scroll_x,
            scroll_y=scroll_y
        )
        self._captured_event_info: EventInfo | None = None

    def _capture(
        self: Self,
        event_info: EventInfo
    ) -> bool:
        target = self._event_info
        if (
            target.event_type == event_info.event_type
            and all(
                target_value is None or source_value is None
                or target_value == (target_value & source_value if masked else source_value)
                for target_value, source_value, masked in (
                    (target.symbol, event_info.symbol, False),
                    (target.modifiers, event_info.modifiers, True),
                    (target.buttons, event_info.buttons, True),
                )
            )
        ):
            self._captured_event_info = event_info
            return True
        return False

    def captured(
        self: Self
    ) -> bool:
        return self._captured_event_info is not None or Toplevel._get_window().capture_event_info_by(self)

    def get_captured_event_info(
        self: Self
    ) -> EventInfo | None:
        return self._captured_event_info


class KeyPress(Event):
    __slots__ = ()

    def __init__(
        self: Self,
        symbol: int | None = None,
        modifiers: int | None = None
    ) -> None:
        super().__init__(
            event_type="key_press",
            symbol=symbol,
            modifiers=modifiers
        )


class KeyRelease(Event):
    __slots__ = ()

    def __init__(
        self: Self,
        symbol: int | None = None,
        modifiers: int | None = None
    ) -> None:
        super().__init__(
            event_type="key_release",
            symbol=symbol,
            modifiers=modifiers
        )


class MouseMotion(Event):
    __slots__ = ()

    def __init__(
        self: Self,
        x: int | None = None,
        y: int | None = None,
        dx: int | None = None,
        dy: int | None = None
    ) -> None:
        super().__init__(
            event_type="mouse_motion",
            x=x,
            y=y,
            dx=dx,
            dy=dy
        )


class MouseDrag(Event):
    __slots__ = ()

    def __init__(
        self: Self,
        x: int | None = None,
        y: int | None = None,
        dx: int | None = None,
        dy: int | None = None,
        buttons: int | None = None,
        modifiers: int | None = None
    ) -> None:
        super().__init__(
            event_type="mouse_drag",
            x=x,
            y=y,
            dx=dx,
            dy=dy,
            buttons=buttons,
            modifiers=modifiers
        )


class MousePress(Event):
    __slots__ = ()

    def __init__(
        self: Self,
        x: int | None = None,
        y: int | None = None,
        buttons: int | None = None,
        modifiers: int | None = None
    ) -> None:
        super().__init__(
            event_type="mouse_press",
            x=x,
            y=y,
            buttons=buttons,
            modifiers=modifiers
        )


class MouseRelease(Event):
    __slots__ = ()

    def __init__(
        self: Self,
        x: int | None = None,
        y: int | None = None,
        buttons: int | None = None,
        modifiers: int | None = None
    ) -> None:
        super().__init__(
            event_type="mouse_release",
            x=x,
            y=y,
            buttons=buttons,
            modifiers=modifiers
        )


class MouseScroll(Event):
    __slots__ = ()

    def __init__(
        self: Self,
        x: int | None = None,
        y: int | None = None,
        scroll_x: float | None = None,
        scroll_y: float | None = None
    ) -> None:
        super().__init__(
            event_type="mouse_scroll",
            x=x,
            y=y,
            scroll_x=scroll_x,
            scroll_y=scroll_y
        )



from __future__ import annotations


from typing import Self

from ...animatables.animatable.piecewiser import Piecewiser
from ...mobjects.mobject import Mobject
from ..timeline import Timeline


class FlashPiecewiser(Piecewiser):
    __slots__ = ("_proportion",)

    def __init__(
        self: Self,
        proportion: float,
        n_segments: int,
        backwards: bool
    ) -> None:
        super().__init__(
            n_segments=n_segments,
            backwards=backwards
        )
        self._proportion: float = proportion

    def get_segment(
        self: Self,
        alpha: float
    ) -> tuple[float, float]:
        proportion = self._proportion
        return (min(alpha * (1.0 + proportion), 1.0), max(alpha * (1.0 + proportion) - proportion, 0.0))


class Flash(Timeline):
    __slots__ = (
        "_mobject",
        "_piecewiser"
    )

    def __init__(
        self: Self,
        mobject: Mobject,
        *,
        proportion: float = 1.0 / 16,
        n_segments: int = 1,
        backwards: bool = False
    ) -> None:
        super().__init__(run_alpha=1.0)
        self._mobject: Mobject = mobject
        self._piecewiser: FlashPiecewiser = FlashPiecewiser(
            proportion=proportion,
            n_segments=n_segments,
            backwards=backwards
        )

    async def construct(
        self: Self
    ) -> None:
        mobject = self._mobject

        self.scene.add(mobject)
        await self.play(mobject.animate().piecewise(mobject.copy(), self._piecewiser))
        self.scene.discard(mobject)



from __future__ import annotations


from typing import Self

import numpy as np

from ...constants.custom_typing import NP_f8
from .animatable_array import AnimatableArray


class AnimatableFloat(AnimatableArray[NP_f8]):
    __slots__ = ()

    def __init__(
        self: Self,
        value: float
    ) -> None:
        super().__init__(np.asarray(value, dtype=np.float64))



from __future__ import annotations


from typing import Self

from ...constants.custom_typing import ShapeType
from ...lazy.lazy import Lazy
from ...lazy.lazy_object import LazyObject


class Buffer(LazyObject):
    __slots__ = ()

    def __init__(
        self: Self,
        shape: ShapeType | None = None,
        array_lens: dict[str, int] | None = None
    ) -> None:
        super().__init__()
        if shape is not None:
            self._shape_ = shape
        if array_lens is not None:
            self._array_len_items_ = tuple(array_lens.items())

    @Lazy.variable()
    @staticmethod
    def _shape_() -> ShapeType:
        return ()

    @Lazy.variable(plural=True)
    @staticmethod
    def _array_len_items_() -> tuple[tuple[str, int], ...]:
        return ()

    @Lazy.property(plural=True)
    @staticmethod
    def _macros_(
        array_len_items: tuple[tuple[str, int], ...]
    ) -> tuple[str, ...]:
        return tuple(
            f"#define {array_len_name} {array_len}"
            for array_len_name, array_len in array_len_items
        )



from __future__ import annotations


from typing import Self

from ...mobjects.mobject import Mobject
from ..timeline import Timeline


class Transform(Timeline):
    __slots__ = (
        "_start_mobject",
        "_stop_mobject"
    )

    def __init__(
        self: Self,
        start_mobject: Mobject,
        stop_mobject: Mobject
    ) -> None:
        super().__init__(run_alpha=1.0)
        self._start_mobject: Mobject = start_mobject
        self._stop_mobject: Mobject = stop_mobject

    async def construct(
        self: Self
    ) -> None:
        start_mobject = self._start_mobject
        stop_mobject = self._stop_mobject
        intermediate_mobject = start_mobject.copy()

        self.scene.discard(start_mobject)
        self.scene.add(intermediate_mobject)
        await self.play(intermediate_mobject.animate().interpolate(start_mobject, stop_mobject))
        self.scene.discard(intermediate_mobject)
        self.scene.add(stop_mobject)



from __future__ import annotations


from typing import (
    Iterator,
    Self,
    Unpack
)

import numpy as np

from ...lazy.lazy import Lazy
from ..animatable.actions import Action
from ..animatable.animatable import (
    Animatable,
    AnimatableActions,
    AnimatableInterpolateAnimation,
    DynamicAnimatable
)
from ..animatable.animation import (
    AnimateKwargs,
    Animation
)


class AnimatableArrayActions(AnimatableActions):
    __slots__ = ()

    @Action.register()
    @classmethod
    def interpolate(
        cls: type[Self],
        dst: AnimatableArray,
        src_0: AnimatableArray,
        src_1: AnimatableArray
    ) -> Iterator[Animation]:
        yield AnimatableArrayInterpolateAnimation(dst, src_0, src_1)


class AnimatableArray[NDArrayT: np.ndarray](AnimatableArrayActions, Animatable):
    __slots__ = ()

    def __init__(
        self: Self,
        array: NDArrayT | None = None
    ) -> None:
        super().__init__()
        if array is not None:
            self._array_ = array

    @Lazy.variable()
    @staticmethod
    def _array_() -> NDArrayT:
        return NotImplemented

    def animate(
        self: Self,
        **kwargs: Unpack[AnimateKwargs]
    ) -> DynamicAnimatableArray[Self]:
        return DynamicAnimatableArray(self, **kwargs)


class DynamicAnimatableArray[AnimatableArrayT: AnimatableArray](AnimatableArrayActions, DynamicAnimatable[AnimatableArrayT]):
    __slots__ = ()


class AnimatableArrayInterpolateAnimation[AnimatableArrayT: AnimatableArray](AnimatableInterpolateAnimation[AnimatableArrayT]):
    __slots__ = ()

    def interpolate(
        self: Self,
        dst: AnimatableArrayT,
        alpha: float
    ) -> None:
        dst._array_ = (1.0 - alpha) * self._src_0_._array_ + alpha * self._src_1_._array_



from __future__ import annotations


from typing import (
    Self,
    Unpack
)

import attrs

from ...constants.custom_typing import AlignmentType
from ...toplevel.toplevel import Toplevel
from .string_mobject import StringMobject
from .tex import (
    TexIO,
    TexInput,
    TexKwargs
)


@attrs.frozen(kw_only=True)
class MathTexInput(TexInput):
    alignment: AlignmentType = attrs.field(factory=lambda: Toplevel._get_config().math_tex_alignment)
    inline: bool = attrs.field(factory=lambda: Toplevel._get_config().math_tex_inline)


class MathTexKwargs(TexKwargs, total=False):
    inline: bool


class MathTexIO[MathTexInputT: MathTexInput](TexIO[MathTexInputT]):
    __slots__ = ()

    @classmethod
    def _get_subdir_name(
        cls: type[Self]
    ) -> str:
        return "math_tex"

    @classmethod
    def _get_environment_command_pair(
        cls: type[Self],
        input_data: MathTexInputT
    ) -> tuple[str, str]:
        if input_data.inline:
            return "$", "$"
        return "\\begin{align*}\n", "\n\\end{align*}"


class MathTex(StringMobject):
    __slots__ = ()

    def __init__(
        self: Self,
        string: str,
        **kwargs: Unpack[MathTexKwargs]
    ) -> None:
        super().__init__(MathTexIO.get(MathTexInput(string=string, **kwargs)))



from __future__ import annotations


__version__ = "0.1.0-alpha"

from .animatables.lights.ambient_light import AmbientLight
from .animatables.lights.point_light import PointLight
from .animatables.camera import Camera
from .animatables.graph import Graph
from .animatables.lighting import Lighting
from .animatables.mesh import Mesh
from .animatables.point import Point
from .animatables.shape import Shape

from .constants.constants import *
from .constants.custom_typing import *
from .constants.palette import *
from .constants.pyglet_constants import *
from .constants.rates import Rates

from .lazy.lazy import Lazy
from .lazy.lazy_object import LazyObject

from .mobjects.graph_mobjects.graph_mobject import GraphMobject
from .mobjects.graph_mobjects.line import Line
from .mobjects.graph_mobjects.polyline import Polyline
from .mobjects.mesh_mobjects.parametric_surface import ParametricSurface
from .mobjects.mesh_mobjects.plane import Plane
from .mobjects.mesh_mobjects.sphere import Sphere
from .mobjects.mesh_mobjects.mesh_mobject import MeshMobject
from .mobjects.shape_mobjects.circle import Circle
from .mobjects.shape_mobjects.polygon import Polygon
from .mobjects.shape_mobjects.polyhedra import (
    Cube,
    Dodecahedron,
    Icosahedron,
    Octahedron,
    Polyhedron,
    Tetrahedron
)
from .mobjects.shape_mobjects.regular_polygon import RegularPolygon
from .mobjects.shape_mobjects.shape_mobject import ShapeMobject
from .mobjects.shape_mobjects.square import Square
from .mobjects.string_mobjects.code import Code
from .mobjects.string_mobjects.markup import Markup
from .mobjects.string_mobjects.math_tex import MathTex
from .mobjects.string_mobjects.mathjax import MathJax
from .mobjects.string_mobjects.tex import Tex
from .mobjects.string_mobjects.text import Text
from .mobjects.image_mobject import ImageMobject
from .mobjects.mobject import Mobject
from .mobjects.svg_mobject import SVGMobject

from .rendering.buffers.attributes_buffer import AttributesBuffer
from .rendering.buffers.texture_buffer import TextureBuffer
from .rendering.buffers.uniform_block_buffer import UniformBlockBuffer
from .rendering.mgl_enums import (
    BlendEquation,
    BlendFunc,
    ContextFlag,
    PrimitiveMode
)
from .rendering.vertex_array import VertexArray

from .timelines.composition.lagged import Lagged
from .timelines.composition.parallel import Parallel
from .timelines.composition.series import Series
from .timelines.fade.fade_in import FadeIn
from .timelines.fade.fade_out import FadeOut
from .timelines.fade.fade_transform import FadeTransform
from .timelines.piecewise.create import Create
from .timelines.piecewise.dashed import Dashed
from .timelines.piecewise.flash import Flash
from .timelines.piecewise.uncreate import Uncreate
from .timelines.transform.transform import Transform
from .timelines.transform.transform_matching_strings import TransformMatchingStrings
from .timelines.timeline import Timeline

from .toplevel.config import Config
from .toplevel.events import (
    KeyPress,
    KeyRelease,
    MouseDrag,
    MouseMotion,
    MousePress,
    MouseRelease,
    MouseScroll
)
from .toplevel.scene import Scene
from .toplevel.toplevel import Toplevel



from __future__ import annotations


from typing import (
    Self,
    Unpack
)

import attrs

from .pango_string_mobject import (
    PangoStringMobjectIO,
    PangoStringMobjectInput,
    PangoStringMobjectKwargs
)
from .string_mobject import StringMobject


@attrs.frozen(kw_only=True)
class TextInput(PangoStringMobjectInput):
    pass


class TextKwargs(PangoStringMobjectKwargs, total=False):
    pass


class TextIO[TextInputT: TextInput](PangoStringMobjectIO[TextInputT]):
    __slots__ = ()

    @classmethod
    def _get_subdir_name(
        cls: type[Self]
    ) -> str:
        return "text"


class Text(StringMobject):
    __slots__ = ()

    def __init__(
        self: Self,
        string: str,
        **kwargs: Unpack[TextKwargs]
    ) -> None:
        super().__init__(TextIO.get(TextInput(string=string, **kwargs)))



from __future__ import annotations


from typing import Self

import numpy as np

from ..constants.constants import ORIGIN
from ..constants.custom_typing import (
    NP_3f8,
    NP_x3f8
)
from ..lazy.lazy import Lazy
from .model import Model


class Point(Model):
    __slots__ = ()

    def __init__(
        self: Self,
        position: NP_3f8 | None = None
    ) -> None:
        super().__init__()
        if position is not None:
            self.shift(position)

    @Lazy.property()
    @staticmethod
    def _local_sample_positions_() -> NP_x3f8:
        return np.array((ORIGIN,))



from __future__ import annotations


from typing import Self

import numpy as np
from PIL import Image

from ..toplevel.toplevel import Toplevel
from .mesh_mobjects.plane import Plane


class ImageMobject(Plane):
    __slots__ = ()

    def __init__(
        self: Self,
        image_filename: str,
        *,
        width: float | None = None,
        height: float | None = 4.0,
        scale: float | None = None
    ) -> None:
        super().__init__()
        for image_dir in Toplevel._get_config().image_search_dirs:
            if (image_path := image_dir.joinpath(image_filename)).exists():
                break
        else:
            raise FileNotFoundError(image_filename)

        image = Image.open(image_path).transpose(Image.Transpose.FLIP_TOP_BOTTOM)
        image_texture = Toplevel._get_context().texture(size=image.size, components=3, samples=0, dtype="f1")
        image_texture.write(image.tobytes("raw", "RGB"))
        self._color_maps_ = (image_texture,)

        pixel_per_unit = Toplevel._get_config().pixel_per_unit
        original_width = image.width / pixel_per_unit
        original_height = image.height / pixel_per_unit
        scale_x, scale_y = type(self)._get_scale_vector(
            original_width=original_width,
            original_height=original_height,
            specified_width=width,
            specified_height=height,
            specified_scale=scale
        )
        self.scale(np.array((
            scale_x * original_width / 2.0,
            scale_y * original_height / 2.0,
            1.0
        )))

    @classmethod
    def _get_scale_vector(
        cls: type[Self],
        *,
        original_width: float,
        original_height: float,
        specified_width: float | None,
        specified_height: float | None,
        specified_scale: float | None
    ) -> tuple[float, float]:
        match specified_width, specified_height:
            case float(), float():
                return specified_width / original_width, specified_height / original_height
            case float(), None:
                scale = specified_width / original_width
            case None, float():
                scale = specified_height / original_height
            case _:
                scale = specified_scale if specified_scale is not None else 1.0
        return scale, scale



from __future__ import annotations


import collections
import itertools
import pathlib
import re
from abc import abstractmethod
from enum import Enum
from typing import (
    Iterable,
    Iterator,
    Self,
    TypedDict
)

import attrs
import scipy.spatial.distance
import scipy.optimize

from ...animatables.arrays.animatable_color import AnimatableColor
from ...animatables.shape import Shape
from ...constants.custom_typing import (
    ColorType,
    SelectorType
)
from ..shape_mobjects.shape_mobject import ShapeMobject
from ..mobject import Mobject
from ..mobject_io import (
    MobjectIO,
    MobjectInput,
    MobjectJSON,
    MobjectOutput
)
from ..svg_mobject import (
    SVGMobjectIO,
    ShapeMobjectJSON
)


class BoundaryFlag(Enum):
    START = 1
    STOP = -1

    def negate(
        self: Self
    ) -> BoundaryFlag:
        return BoundaryFlag(-self.value)


class Span:
    __slots__ = (
        "_start",
        "_stop"
    )

    def __init__(
        self: Self,
        start: int,
        stop: int
    ) -> None:
        assert 0 <= start <= stop
        super().__init__()
        self._start: int = start
        self._stop: int = stop

    def contains(
        self: Self,
        span: Span
    ) -> bool:
        return self._start <= span._start and self._stop >= span._stop

    def as_slice(
        self: Self
    ) -> slice:
        return slice(self._start, self._stop)

    def get_boundary_index(
        self: Self,
        boundary_flag: BoundaryFlag
    ) -> int:
        return self._start if boundary_flag == BoundaryFlag.START else self._stop


class CommandFlag(Enum):
    STANDALONE = 0
    OPEN = 1
    CLOSE = -1


class CommandInfo[AttributesT: TypedDict]:
    __slots__ = (
        "_command_items",
        "_attributes_item"
    )

    def __init__(
        self: Self,
        match_obj_items: tuple[tuple[re.Match[str], str | None, CommandFlag], ...],
        attributes_item: tuple[AttributesT, bool] | None
    ) -> None:
        super().__init__()
        self._command_items: tuple[tuple[Span, str, CommandFlag], ...] = tuple(
            (
                Span(*match.span()),
                replacement if replacement is not None else match.group(),
                command_flag
            )
            for match, replacement, command_flag in match_obj_items
        )
        self._attributes_item: tuple[AttributesT, bool] | None = attributes_item


class StandaloneCommandInfo[AttributesT: TypedDict](CommandInfo[AttributesT]):
    __slots__ = ()

    def __init__(
        self: Self,
        match: re.Match[str],
        replacement: str | None = None
    ) -> None:
        super().__init__(
            match_obj_items=(
                (match, replacement, CommandFlag.STANDALONE),
            ),
            attributes_item=None
        )


class BalancedCommandInfo[AttributesT: TypedDict](CommandInfo[AttributesT]):
    __slots__ = (
        "_other_match_obj_item",
        "_attributes_item"
    )

    def __init__(
        self: Self,
        attributes: AttributesT,
        isolated: bool,
        open_match_obj: re.Match[str],
        close_match_obj: re.Match[str],
        open_replacement: str | None = None,
        close_replacement: str | None = None
    ) -> None:
        super().__init__(
            match_obj_items=(
                (open_match_obj, open_replacement, CommandFlag.OPEN),
                (close_match_obj, close_replacement, CommandFlag.CLOSE)
            ),
            attributes_item=(attributes, isolated)
        )


@attrs.frozen(kw_only=True)
class SpanInfo[AttributesT: TypedDict]:
    span: Span
    isolated: bool | None = None
    attributes: AttributesT | None = None
    local_color: ColorType | None = None
    command_item: tuple[CommandInfo[AttributesT], str, CommandFlag] | None = None


class ReplacementRecord:
    __slots__ = (
        "_span",
        "_unlabelled_replacement",
        "_labelled_replacement"
    )

    def __init__(
        self: Self,
        span: Span
    ) -> None:
        super().__init__()
        self._span: Span = span
        self._unlabelled_replacement: str = ""
        self._labelled_replacement: str = ""

    def write_replacements(
        self: Self,
        *,
        unlabelled_replacement: str,
        labelled_replacement: str
    ) -> Self:
        self._unlabelled_replacement = unlabelled_replacement
        self._labelled_replacement = labelled_replacement
        return self


class InsertionRecord(ReplacementRecord):
    __slots__ = ()

    def __init__(
        self: Self,
        index: int
    ) -> None:
        super().__init__(Span(index, index))


class StringMobjectKwargs[AttributesT: TypedDict](TypedDict, total=False):
    global_attributes: AttributesT
    local_attributes: dict[SelectorType, AttributesT]
    local_colors: dict[SelectorType, ColorType]
    isolate: Iterable[SelectorType]
    protect: Iterable[SelectorType]
    concatenate: bool


@attrs.frozen(kw_only=True)
class StringMobjectInput[AttributesT: TypedDict](MobjectInput):
    string: str
    global_attributes: AttributesT = attrs.field(factory=dict)
    local_attributes: dict[SelectorType, AttributesT] = attrs.field(factory=dict)
    local_colors: dict[SelectorType, ColorType] = attrs.field(factory=dict)
    isolate: Iterable[SelectorType] = ()
    protect: Iterable[SelectorType] = ()
    concatenate: bool = False


@attrs.frozen(kw_only=True)
class StringMobjectOutput(MobjectOutput):
    shape_mobjects: tuple[ShapeMobject, ...]
    string: str
    labels: tuple[int, ...]
    isolated_spans: tuple[Span, ...]


class StringMobjectJSON(MobjectJSON):
    shape_mobjects: tuple[ShapeMobjectJSON, ...]
    string: str
    labels: tuple[int, ...]
    isolated_spans: tuple[tuple[int, ...], ...]


class StringMobjectIO[AttributesT: TypedDict, StringMobjectInputT: StringMobjectInput](
    MobjectIO[StringMobjectInputT, StringMobjectOutput, StringMobjectJSON]
):
    __slots__ = ()

    @classmethod
    def generate(
        cls: type[Self],
        input_data: StringMobjectInputT,
        temp_path: pathlib.Path
    ) -> StringMobjectOutput:

        def replace_string(
            original_pieces: tuple[str, ...],
            replacement_pieces: tuple[str, ...]
        ) -> str:
            return "".join(itertools.chain.from_iterable(zip(
                ("", *original_pieces),
                replacement_pieces,
                strict=True
            )))

        string = input_data.string
        isolated_items, replacement_records = cls._get_isolated_items_and_replacement_records(
            string=string,
            environment_command_pair=cls._get_environment_command_pair(input_data),
            global_attributes_items=tuple(cls._iter_global_span_attributes(input_data, temp_path)),
            local_span_infos=tuple(itertools.chain((
                SpanInfo(
                    span=span,
                    isolated=False,
                    attributes=attributes
                )
                for span, attributes in cls._iter_local_span_attributes(input_data, temp_path)
            ), (
                SpanInfo(
                    span=span,
                    isolated=True,
                    attributes=cls._get_empty_attributes(),
                    local_color=local_color
                )
                for selector, local_color in input_data.local_colors.items()
                for span in cls._iter_spans_by_selector(selector, string)
            ), (
                SpanInfo(
                    span=span,
                    isolated=True,
                    attributes=cls._get_empty_attributes()
                )
                for selector in input_data.isolate
                for span in cls._iter_spans_by_selector(selector, string)
            ), (
                SpanInfo(
                    span=span
                )
                for selector in input_data.protect
                for span in cls._iter_spans_by_selector(selector, string)
            ), (
                SpanInfo(
                    span=span,
                    command_item=(command_info, replacement, command_flag)
                )
                for command_info in cls._iter_command_infos(string)
                for span, replacement, command_flag in command_info._command_items
            )))
        )
        original_pieces = tuple(
            string[start:stop]
            for start, stop in zip(
                (replacement_record._span._stop for replacement_record in replacement_records[:-1]),
                (replacement_record._span._start for replacement_record in replacement_records[1:]),
                strict=True
            )
        )
        shape_mobject_items = tuple(cls._iter_shape_mobject_items(
            unlabelled_content=replace_string(
                original_pieces=original_pieces,
                replacement_pieces=tuple(
                    replacement_record._unlabelled_replacement
                    for replacement_record in replacement_records
                )
            ),
            labelled_content=replace_string(
                original_pieces=original_pieces,
                replacement_pieces=tuple(
                    replacement_record._labelled_replacement
                    for replacement_record in replacement_records
                )
            ),
            requires_labelling=len(isolated_items) > 1,
            input_data=input_data,
            temp_path=temp_path
        ))

        for shape_mobject, label in shape_mobject_items:
            _, local_color = isolated_items[label]
            if local_color is not None:
                shape_mobject._color_._array_ = AnimatableColor._color_to_array(local_color)

        return StringMobjectOutput(
            shape_mobjects=tuple(shape_mobject for shape_mobject, _ in shape_mobject_items),
            string=string,
            labels=tuple(label for _, label in shape_mobject_items),
            isolated_spans=tuple(isolated_span for isolated_span, _ in isolated_items)
        )

    @classmethod
    def dump_json(
        cls: type[Self],
        output_data: StringMobjectOutput
    ) -> StringMobjectJSON:
        return StringMobjectJSON(
            shape_mobjects=tuple(
                SVGMobjectIO._shape_mobject_to_json(shape_mobject)
                for shape_mobject in output_data.shape_mobjects
            ),
            string=output_data.string,
            labels=output_data.labels,
            isolated_spans=tuple((span._start, span._stop) for span in output_data.isolated_spans)
        )

    @classmethod
    def load_json(
        cls: type[Self],
        json_data: StringMobjectJSON
    ) -> StringMobjectOutput:
        return StringMobjectOutput(
            shape_mobjects=tuple(
                SVGMobjectIO._json_to_shape_mobject(shape_mobject_json)
                for shape_mobject_json in json_data["shape_mobjects"]
            ),
            string=json_data["string"],
            labels=json_data["labels"],
            isolated_spans=tuple(Span(*span_values) for span_values in json_data["isolated_spans"])
        )

    @classmethod
    def _get_isolated_items_and_replacement_records(
        cls: type[Self],
        string: str,
        environment_command_pair: tuple[str, str],
        global_attributes_items: tuple[AttributesT, ...],
        local_span_infos: tuple[SpanInfo[AttributesT], ...]
    ) -> tuple[tuple[tuple[Span, ColorType | None], ...], tuple[ReplacementRecord, ...]]:

        def get_sorting_key(
            span_boundary: tuple[SpanInfo[AttributesT], BoundaryFlag]
        ) -> tuple[int, int, int]:
            span_info, boundary_flag = span_boundary
            span = span_info.span
            index = span.get_boundary_index(boundary_flag)
            paired_index = span.get_boundary_index(boundary_flag.negate())
            # All local spans are guaranteed to have non-zero widths.
            return (
                index,
                boundary_flag.value,
                -paired_index
            )

        insertion_record_items: list[tuple[InsertionRecord, InsertionRecord, AttributesT, bool, ColorType | None]] = []
        replacement_records: collections.deque[ReplacementRecord] = collections.deque()
        bracket_counter = itertools.count()
        protect_level = 0
        bracket_stack: list[int] = []
        open_command_stack: list[tuple[InsertionRecord, CommandInfo[AttributesT]]] = []
        start_stack: list[tuple[SpanInfo[AttributesT], InsertionRecord, int, tuple[int, ...]]] = []
        local_color_stack: list[ColorType] = []

        for span_info, boundary_flag in sorted((
            (span_info, boundary_flag)
            for boundary_flag in (BoundaryFlag.STOP, BoundaryFlag.START)
            for span_info in local_span_infos[::boundary_flag.value]
        ), key=get_sorting_key):
            span = span_info.span
            if span_info.isolated is None:
                protect_level += boundary_flag.value
                if span_info.command_item is None:
                    continue
                if boundary_flag == BoundaryFlag.START:
                    continue
                command_info, command_replacement, command_flag = span_info.command_item
                command_replacement_record = ReplacementRecord(span).write_replacements(
                    unlabelled_replacement=command_replacement,
                    labelled_replacement=command_replacement
                )
                if command_flag == CommandFlag.OPEN:
                    bracket_stack.append(next(bracket_counter))
                    open_insertion_record = InsertionRecord(span._stop)
                    replacement_records.append(command_replacement_record)
                    replacement_records.append(open_insertion_record)
                    open_command_stack.append((open_insertion_record, command_info))
                elif command_flag == CommandFlag.CLOSE:
                    bracket_stack.pop()
                    close_insertion_record = InsertionRecord(span._start)
                    replacement_records.append(close_insertion_record)
                    replacement_records.append(command_replacement_record)
                    open_insertion_record, open_command_info = open_command_stack.pop()
                    assert open_command_info is command_info
                    assert command_info._attributes_item is not None
                    attributes, isolated = command_info._attributes_item
                    insertion_record_items.append((
                        open_insertion_record,
                        close_insertion_record,
                        attributes,
                        isolated,
                        None
                    ))
                elif command_flag == CommandFlag.STANDALONE:
                    replacement_records.append(command_replacement_record)
                continue

            if boundary_flag == BoundaryFlag.START:
                start_insertion_record = InsertionRecord(span._start)
                replacement_records.append(start_insertion_record)
                start_stack.append((span_info, start_insertion_record, protect_level, tuple(bracket_stack)))
                if span_info.local_color is not None:
                    local_color_stack.append(span_info.local_color)
            elif boundary_flag == BoundaryFlag.STOP:
                stop_insertion_record = InsertionRecord(span._stop)
                replacement_records.append(stop_insertion_record)
                start_span_info, start_insertion_record, start_protect_level, start_bracket_stack = start_stack.pop()
                if start_protect_level or protect_level:
                    continue
                if start_span_info is not span_info:
                    raise ValueError(
                        f"Partly overlapping substrings detected: '{string[start_span_info.span.as_slice()]}', '{string[span.as_slice()]}'"
                    )
                if start_bracket_stack != tuple(bracket_stack):
                    raise ValueError(
                        f"Cannot handle substring: '{string[span.as_slice()]}'"
                    )
                if span_info.local_color is not None:
                    local_color = span_info.local_color
                    local_color_stack.pop()
                else:
                    local_color = local_color_stack[-1] if local_color_stack else None
                assert span_info.attributes is not None
                insertion_record_items.append((
                    start_insertion_record,
                    stop_insertion_record,
                    span_info.attributes,
                    span_info.isolated,
                    local_color
                ))

        assert protect_level == 0
        assert not bracket_stack
        assert not open_command_stack
        assert not start_stack
        assert not local_color_stack

        label_counter = itertools.count()
        isolated_items: list[tuple[Span, ColorType | None]] = []

        global_label = next(label_counter)
        global_span = Span(0, len(string))
        isolated_items.append((global_span, None))

        start_environment_command, stop_environment_command = environment_command_pair
        replacement_records.appendleft(InsertionRecord(global_span._start).write_replacements(
            unlabelled_replacement=start_environment_command,
            labelled_replacement=start_environment_command
        ))
        replacement_records.append(InsertionRecord(global_span._stop).write_replacements(
            unlabelled_replacement=stop_environment_command,
            labelled_replacement=stop_environment_command
        ))

        for attributes in reversed(global_attributes_items):
            labelled_attributes = cls._convert_attributes_for_labelling(attributes, global_label)
            start_unlabelled_insertion, stop_unlabelled_insertion = cls._get_command_pair(attributes)
            start_labelled_insertion, stop_labelled_insertion = cls._get_command_pair(labelled_attributes)
            replacement_records.appendleft(InsertionRecord(global_span._start).write_replacements(
                unlabelled_replacement=start_unlabelled_insertion,
                labelled_replacement=start_labelled_insertion
            ))
            replacement_records.append(InsertionRecord(global_span._stop).write_replacements(
                unlabelled_replacement=stop_unlabelled_insertion,
                labelled_replacement=stop_labelled_insertion
            ))

        for start_insertion_record, stop_insertion_record, attributes, isolated, local_color in insertion_record_items:
            if isolated:
                local_label = next(label_counter)
                local_span = Span(start_insertion_record._span._stop, stop_insertion_record._span._start)
                isolated_items.append((local_span, local_color))
            else:
                local_label = None
            labelled_attributes = cls._convert_attributes_for_labelling(attributes, local_label)
            start_unlabelled_insertion, stop_unlabelled_insertion = cls._get_command_pair(attributes)
            start_labelled_insertion, stop_labelled_insertion = cls._get_command_pair(labelled_attributes)
            start_insertion_record.write_replacements(
                unlabelled_replacement=start_unlabelled_insertion,
                labelled_replacement=start_labelled_insertion
            )
            stop_insertion_record.write_replacements(
                unlabelled_replacement=stop_unlabelled_insertion,
                labelled_replacement=stop_labelled_insertion
            )

        return tuple(isolated_items), tuple(replacement_records)

    @classmethod
    def _iter_spans_by_selector(
        cls: type[Self],
        selector: SelectorType,
        string: str
    ) -> Iterator[Span]:
        match selector:
            case str():
                substr_len = len(selector)
                if not substr_len:
                    return
                index = 0
                while True:
                    index = string.find(selector, index)
                    if index == -1:
                        break
                    yield Span(index, index + substr_len)
                    index += substr_len
            case re.Pattern():
                for match in selector.finditer(string):
                    start, stop = match.span()
                    if start < stop:
                        yield Span(start, stop)
            case slice(start=int(start), stop=int(stop)):
                l = len(string)
                start = min(start, l) if start >= 0 else max(start + l, 0)
                stop = min(stop, l) if stop >= 0 else max(stop + l, 0)
                if start < stop:
                    yield Span(start, stop)

    @classmethod
    def _iter_shape_mobject_items(
        cls: type[Self],
        unlabelled_content: str,
        labelled_content: str,
        requires_labelling: bool,
        input_data: StringMobjectInputT,
        temp_path: pathlib.Path
    ) -> Iterator[tuple[ShapeMobject, int]]:
        unlabelled_shape_mobjects = cls._get_shape_mobjects(unlabelled_content, input_data, temp_path)
        if input_data.concatenate:
            yield ShapeMobject(Shape().concatenate(tuple(
                shape_mobject._shape_ for shape_mobject in unlabelled_shape_mobjects
            ))), 0
            return

        if not requires_labelling or not unlabelled_shape_mobjects:
            for unlabelled_shape_mobject in unlabelled_shape_mobjects:
                yield unlabelled_shape_mobject, 0
            return

        labelled_shape_mobjects = cls._get_shape_mobjects(labelled_content, input_data, temp_path)
        assert len(unlabelled_shape_mobjects) == len(labelled_shape_mobjects)

        unlabelled_radii = Mobject().add(*unlabelled_shape_mobjects).box.get_radii()
        labelled_radii = Mobject().add(*labelled_shape_mobjects).box.get_radii()
        scale_factor = unlabelled_radii / labelled_radii
        distance_matrix = scipy.spatial.distance.cdist(
            tuple(shape.box.get() for shape in unlabelled_shape_mobjects),
            tuple(shape.box.get() * scale_factor for shape in labelled_shape_mobjects)
        )
        for unlabelled_index, labelled_index in zip(*scipy.optimize.linear_sum_assignment(distance_matrix), strict=True):
            yield (
                unlabelled_shape_mobjects[unlabelled_index],
                int(AnimatableColor._array_to_hex(labelled_shape_mobjects[labelled_index]._color_._array_)[1:], 16)
            )

    @classmethod
    def _get_shape_mobjects(
        cls: type[Self],
        content: str,
        input_data: StringMobjectInputT,
        temp_path: pathlib.Path
    ) -> tuple[ShapeMobject, ...]:
        svg_path = temp_path.with_suffix(".svg")
        try:
            cls._create_svg(
                content=content,
                input_data=input_data,
                svg_path=svg_path
            )
            shape_mobjects = SVGMobjectIO._get_shape_mobjects_from_svg_path(
                svg_path=svg_path,
                scale=cls._get_adjustment_scale()
            )
        finally:
            svg_path.unlink(missing_ok=True)

        return shape_mobjects

    @classmethod
    @abstractmethod
    def _create_svg(
        cls: type[Self],
        content: str,
        input_data: StringMobjectInputT,
        svg_path: pathlib.Path
    ) -> None:
        pass

    @classmethod
    @abstractmethod
    def _get_adjustment_scale(
        cls: type[Self]
    ) -> float:
        # The line height shall be roughly equal to 1.0 for default fonts.
        pass

    @classmethod
    def _get_environment_command_pair(
        cls: type[Self],
        input_data: StringMobjectInputT
    ) -> tuple[str, str]:
        return "", ""

    @classmethod
    def _iter_global_span_attributes(
        cls: type[Self],
        input_data: StringMobjectInputT,
        temp_path: pathlib.Path
    ) -> Iterator[AttributesT]:
        yield input_data.global_attributes

    @classmethod
    def _iter_local_span_attributes(
        cls: type[Self],
        input_data: StringMobjectInputT,
        temp_path: pathlib.Path
    ) -> Iterator[tuple[Span, AttributesT]]:
        for selector, attributes in input_data.local_attributes.items():
            for span in cls._iter_spans_by_selector(selector, input_data.string):
                yield span, attributes

    @classmethod
    @abstractmethod
    def _get_empty_attributes(
        cls: type[Self]
    ) -> AttributesT:
        pass

    @classmethod
    @abstractmethod
    def _get_command_pair(
        cls: type[Self],
        attributes: AttributesT
    ) -> tuple[str, str]:
        pass

    @classmethod
    @abstractmethod
    def _convert_attributes_for_labelling(
        cls: type[Self],
        attributes: AttributesT,
        label: int | None
    ) -> AttributesT:
        pass

    @classmethod
    @abstractmethod
    def _iter_command_infos(
        cls: type[Self],
        string: str
    ) -> Iterator[CommandInfo[AttributesT]]:
        pass


class StringMobject(ShapeMobject):
    """
    This class aims to optimize the logic of "slicing children
    via substrings". This could be much clearer and more user-friendly
    than slicing through numerical indices explicitly.

    Each instance of `StringMobject` generates 2 svg files.
    The additional one is generated with some color commands inserted,
    so that each child of the original svg will be labelled
    by the color of its paired child from the additional svg.
    """
    __slots__ = (
        "_shape_mobjects",
        "_string",
        "_labels",
        "_isolated_spans"
    )

    def __init__(
        self: Self,
        output_data: StringMobjectOutput
    ) -> None:
        super().__init__()

        shape_mobjects = output_data.shape_mobjects
        self._shape_mobjects: tuple[ShapeMobject, ...] = shape_mobjects
        self._string: str = output_data.string
        self._labels: tuple[int, ...] = output_data.labels
        self._isolated_spans: tuple[Span, ...] = output_data.isolated_spans
        self.add(*shape_mobjects)

    def _get_indices_tuple_by_selector(
        self: Self,
        selector: SelectorType
    ) -> tuple[tuple[int, ...], ...]:
        return tuple(
            tuple(
                index
                for index, label in enumerate(self._labels)
                if specified_span.contains(self._isolated_spans[label])
            )
            for specified_span in StringMobjectIO._iter_spans_by_selector(selector, self._string)
        )

    def _build_from_indices(
        self: Self,
        indices: tuple[int, ...]
    ) -> Mobject:
        return Mobject().add(*(
            self._shape_mobjects[index]
            for index in indices
        ))

    def _build_from_indices_tuple(
        self: Self,
        indices_tuple: tuple[tuple[int, ...], ...]
    ) -> Mobject:
        return Mobject().add(*(
            self._build_from_indices(indices)
            for indices in indices_tuple
        ))

    def select_part(
        self: Self,
        selector: SelectorType,
        index: int = 0
    ) -> Mobject:
        return self._build_from_indices(self._get_indices_tuple_by_selector(selector)[index])

    def select_parts(
        self: Self,
        selector: SelectorType
    ) -> Mobject:
        return self._build_from_indices_tuple(self._get_indices_tuple_by_selector(selector))
